
__all__ = ['Requirements', 'WriteOutput']


import os

import numpy as np
import cpf.PeakFunctions as ff
import json
import lmfit
from lmfit.model import load_modelresult

def Requirements():
    #List non-universally required parameters for writing this output type.
    
    RequiredParams = [
            #'apparently none!
            ]
    OptionalParams = [
            'ElasticProperties', #FIX ME: this needs to be included
            'tc', # which thermocoule to include from 6BMB/X17B2 collection system. default to 1. #FIX ME: this needs to be included
            'Output_directory' # if no direcrtory is specified write to current directory.
            ]
    
    return RequiredParams
    

def lmfit_fix_int_data_type(fname):
    """
    fixes problem with lmfit save/load model. 
    lmfit load model cannot read int32 data with nulls in it. 
    if replace 'int32' with 'float32' it will read. 
    """
            
    ObjRead = open(fname, "r")   
    txtContent = ObjRead.read();
    ObjRead.close()
    
    txtContent = txtContent.replace('int', 'float')
    
    print('    Rewriting', fname)
    ObjRead = open(fname, "w")
    ObjRead.write(txtContent)
    ObjRead.close()
    

    
def WriteOutput(FitSettings, parms_dict):
# writes *.exp files required by polydefixED.
# N.B. this is a different file than that required by polydefix for monochromatic diffraction.
# This file contains a list of all the diffraction information. Hence it has to be written after the fitting as a single operation.


    FitParameters = dir(FitSettings)
    
    #Parse the required inputs. 
    base_file_name = FitSettings.datafile_Basename
    
    # diffraction patterns 
    
    # Identical to code in XRD_FitPatterns
    if not 'datafile_Files' in FitParameters:
        n_diff_files = FitSettings.datafile_EndNum - FitSettings.datafile_StartNum + 1
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(j+FitSettings.datafile_StartNum).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    elif 'datafile_Files' in FitParameters:
        n_diff_files = len(FitSettings.datafile_Files)
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(FitSettings.datafile_Files[j]).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    else:
        n_diff_files = len(FitSettings.datafile_Files)

    if 'Output_directory' in FitParameters:
        out_dir = FitSettings.Output_directory
    else:
        out_dir = './'
            

    # create output file name from passed name
    path, filename = os.path.split(base_file_name)
    base, ext = os.path.splitext(filename)
    if not base:
        print("No base filename, using input filename instead.")
        base =  os.path.splitext(FitSettings.inputfile)[0]
        
    if base[-1:] == '_': #if the base file name ends in an '_' remove it. 
        base = base[0:-1]
        
    out_file = out_dir + base + '.dat'

    text_file = open(out_file, "w")
    print('Writing', out_file)
    
    text_file.write("# Summary of fits produced by continuous_peak_fit for input file: %s.\n" % FitSettings.inputfile)
    text_file.write("# For more information: http://www.github.com/me/something\n" )
    text_file.write("# File version: %i \n" % 1 )
    text_file.write("# \n")
    
    #write header
    col_width = 12
    text_file.write(("# {0:<"+str(col_width+5)+"}").format("Data File"+","))
    text_file.write(("{0:<"+str(col_width)+"}").format("Peak"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d0"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d0_err"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d2cos"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d2cos_err"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d2sin"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d2sin_err"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("corr coef"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("diff strain"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("diff s err"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("orientation"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("orient err"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d_max"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("d_min"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("h0"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("h0_err"+",")) 
    text_file.write(("{0:>"+str(col_width)+"}").format("w0"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("w0_err"+",")) 
    text_file.write(("{0:>"+str(col_width)+"}").format("p0"+","))
    text_file.write(("{0:>"+str(col_width)+"}").format("p0_err"+","))
    text_file.write("\n")
    
    for z in range(n_diff_files):
        
        filename = os.path.splitext(os.path.basename(diff_files[z]))[0]
        filename = filename+'.json'
        
        # Read JSON data from file
        with open(filename) as json_data:
            fit = json.load(json_data)
        
        num_subpatterns = len(FitSettings.fit_orders)
        for y in range(num_subpatterns):
            
            orders = FitSettings.fit_orders[y]
            
            subfilename = os.path.splitext(os.path.basename(diff_files[z]))[0] + '_'
            
            for x in range(len(orders['peak'])):
                if 'phase' in orders['peak'][x]:
                    subfilename = subfilename + orders['peak'][x]['phase']
                else:
                    subfilename = subfilename + "Peak"
                if 'hkl' in orders['peak'][x]:
                    subfilename = subfilename + str(orders['peak'][x]['hkl'])
                else:
                    subfilename = subfilename + x
                        
                if x < len(orders['peak']) - 1 and len(orders['peak']) > 1:
                    subfilename = subfilename + '_'
                        
            print('  Incorporating ' + subfilename)
            try:
                gmodel = load_modelresult(subfilename+'.sav', funcdefs={'PeaksModel': ff.PeaksModel})
            except:
                lmfit_fix_int_data_type(subfilename+'.sav')
                try:
                    gmodel = load_modelresult(subfilename+'.sav', funcdefs={'PeaksModel': ff.PeaksModel})
                except:
                    raise FileNotFoundError     
        
            # FIX ME: this will only work for one peak. Needs fixing if more then one peak in subpattern
            corr = gmodel.params['peak_0_d3'].correl['peak_0_d4']
            #ci, trace = lmfit.conf_interval(mini, gmodel, sigmas=[1, 2], trace=True)
            #lmfit.printfuncs.report_ci(ci)

            #get parameters to write to output file.
            #file name
            out_name = os.path.splitext(os.path.basename(diff_files[z]))[0]
            
            for x in range(len(orders['peak'])):
                
                out_peak = []
                
                #peak
                if 'phase' in orders['peak'][x]:
                    out_peak = orders['peak'][x]['phase']
                else:
                    out_peak = out_peak + "Peak"
                if 'hkl' in orders['peak'][x]:
                    out_peak = out_peak + ' (' + str(orders['peak'][x]['hkl']) + ')'
                else:
                    out_peak = out_peak + ' ' + str(x)
                
                # d0
                out_d0       = fit[y]['peak'][x]['d-space'][0]
                out_d0err    = fit[y]['peak'][x]['d-space_err'][0]
                #differential coefficients, errors and covarience
                out_dcos2    = fit[y]['peak'][x]['d-space'][3]
                out_dsin2    = fit[y]['peak'][x]['d-space'][4]
                out_dcos2err = fit[y]['peak'][x]['d-space_err'][3]
                out_dsin2err = fit[y]['peak'][x]['d-space_err'][4]
                out_dcorr    = gmodel.params['peak_0_d3'].correl['peak_0_d4']
                
                #differential strain
                # a.cos(2t) + b.sin(2t) = (a^2+b^2)^(1/2) . cos(2t - atan(b/a))
                # out_dd  = (a^2+b^2)^(1/2)
                out_dd    = (fit[y]['peak'][x]['d-space'][3]**2 + fit[y]['peak'][x]['d-space'][4]**2)**(1/2)
                #out_dderr= [(2.a.da.)^2 + (2.b.db)^2]^(1/2)]^(1/2)
                out_dderr = ((2 * fit[y]['peak'][x]['d-space'][3] * fit[y]['peak'][x]['d-space_err'][3])**2 + 
                             (2 * fit[y]['peak'][x]['d-space'][4] * fit[y]['peak'][x]['d-space_err'][4])**2 ) ** (1/4)
                
                #angle
                # out_ang  = atan(b/a)
                out_ang    = (np.arctan(fit[y]['peak'][x]['d-space'][4] / fit[y]['peak'][x]['d-space'][3]))
                # d (atan(c))/dc = 1/(c^2+1). c = b/a. dc = c.((da/a)^2 + (db/b)^2)^(1/2)
                # out_angerr = dc. 
                out_angerr = 1 / ((fit[y]['peak'][x]['d-space'][4] / fit[y]['peak'][x]['d-space'][3])**2 + 1) * (np.abs(out_ang) * ((fit[y]['peak'][x]['d-space_err'][3]/fit[y]['peak'][x]['d-space'][3])**2 + (fit[y]['peak'][x]['d-space_err'][4]/fit[y]['peak'][x]['d-space'][4])**2)**(1/2))
                # correction to make angle correct (otherwise out by pi)
                if fit[y]['peak'][x]['d-space'][3] < 0:
                    out_ang = out_ang + np.pi
                # force angles to be between -pi and pi
                if out_ang > np.pi/2:
                    out_ang = out_ang - np.pi
                elif out_ang < -np.pi/2:
                    out_ang = out_ang + np.pi
                # convert angles to degrees.
                out_ang    = np.rad2deg(out_ang)
                out_angerr = np.rad2deg(out_angerr)
                
                
                #differential max
                out_dmax    = out_dd * np.cos(2*np.pi/4 - np.deg2rad(out_ang)) + out_d0
                #differential min 
                out_dmin    = out_dd * np.cos(2*3*np.pi/4 - np.deg2rad(out_ang)) + out_d0
                
                #height mean
                out_h0    = fit[y]['peak'][x]['height'][0]
                out_h0err = fit[y]['peak'][x]['height_err'][0]
                #width mean
                out_w0    = fit[y]['peak'][x]['width'][0]
                out_w0err = fit[y]['peak'][x]['width_err'][0]
                #profile mean
                out_p0    = fit[y]['peak'][x]['profile'][0]
                out_p0err = fit[y]['peak'][x]['profile_err'][0]
            
            
            
                #write numbers to file
                dp = 5
                text_file.write(("{0:<"+str(col_width+7)+"}").format(out_name+","))
                text_file.write(("{0:<"+str(col_width)+"}").format(out_peak+","))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_d0))
                #text_file.write(" %10.5f," % out_d0)
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_d0err))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dcos2))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dcos2err))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dsin2))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dsin2err))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dcorr))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dd))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dderr))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_ang))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_angerr))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dmax))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_dmin))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_h0))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_h0err))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_w0))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_w0err))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_p0))
                text_file.write(("{0:"+str(col_width-1)+"."+str(dp)+"f},").format(out_p0err))
                text_file.write("\n")
            
            
            
    # col_width = 10
    # text_file.write(("# {0:<"+str(col_width+5)+"}").format("Data File"+","))
    # text_file.write(("{0:<"+str(col_width)+"}").format("Peak"+","))
    # text_file.write(("{0:<"+str(col_width)+"}").format("d0"+","))
    # text_file.write(("{0:<"+str(col_width)+"}").format("d0_err"+","))
            
    text_file.close()
            
            


            
            
            





__all__ = ['Requirements', 'WriteOutput']


import os

import numpy as np
#import cpf.PeakFunctions as ff
import cpf.WriteMultiFit as WriteMultiFit
import json
#import re
#import datetime

def Requirements():
    #List non-universally required parameters for writing this output type.
    
    RequiredParams = [
            #'apparently none!
            ]
    OptionalParams = [
            'ElasticProperties', #FIX ME: this needs to be included
            'tc', # which thermocoule to include from 6BMB/X17B2 collection system. default to 1. #FIX ME: this needs to be included
            'Output_directory' # if no direcrtory is specified write to current directory.
            ]
    
    return RequiredParams
    
    
def WriteOutput(FitSettings, parms_dict):
# writes *.fit files produced by multifit (as input for polydefix)
# writes *.exp files required by polydefix.
# N.B. this is a different file than that required by polydefix for energy dispersive diffraction.

    
    #Write fit files
    WriteMultiFit.WriteOutput(FitSettings, parms_dict)



    FitParameters = dir(FitSettings)
    
    #Parse the required inputs. 
    base_file_name = FitSettings.datafile_Basename
    
    # diffraction patterns 
    # Identical to code in XRD_FitPatterns
    if not 'datafile_Files' in FitParameters:
        n_diff_files = FitSettings.datafile_EndNum - FitSettings.datafile_StartNum + 1
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(j+FitSettings.datafile_StartNum).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    elif 'datafile_Files' in FitParameters:
        n_diff_files = len(FitSettings.datafile_Files)
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(FitSettings.datafile_Files[j]).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    else:
        n_diff_files = len(FitSettings.datafile_Files)

    if 'Output_directory' in FitParameters:
        out_dir = FitSettings.Output_directory
    else:
        out_dir = './'
            

    # create output file name from passed name
    path, filename = os.path.split(base_file_name)
    base, ext = os.path.splitext(filename)
    if base[-1:] == '_': #if the base file name ends in an '_' remove it. 
        base = base[0:-1]
    out_file = out_dir + base + '.exp'

    text_file = open(out_file, "w")
    print('Writing', out_file)

    #headers. set file version to be 1.
    text_file.write("# Experiment analysis file. to be used with Polydefix\n")
    text_file.write("# For more information: http://merkel.zoneo.net/Polydefix/\n" )
    text_file.write("# File Created by FPF_WritePolydefix function in Continuous-Peak-Fit\n")
    text_file.write("# For more information: http://www.github.com/me/something\n" )
    
    
    text_file.write("# File version\n" )
    text_file.write("     2\n" )

    #Write data properties
    text_file.write("# Directory with FIT files\n" )
    # needs absolute path for the data files and the string has to end with a '/' e.g. Users/me/data/BCC1_2GPa_10s_e/
    text_file.write("     %s/\n" %  os.path.abspath(os.getcwd() + '/' + FitSettings.Output_directory) ) 
    text_file.write("# Basename for FIT files\n" )
    #if last symbol in the file name is '_' then we need to strip it from the name
    text_file.write("     %s\n" %  FitSettings.datafile_Basename.strip('_') )
    text_file.write("# First index for FIT files\n" )
    text_file.write("     %i\n" %  FitSettings.datafile_StartNum )
    text_file.write("# Last index for FIT files\n" )
    text_file.write("     %i\n" %  FitSettings.datafile_EndNum ) 
    text_file.write("# Number of digits for FIT files\n" )
    text_file.write("     %i\n" %  FitSettings.datafile_NumDigit )
    text_file.write("# Wavelength\n" )
    text_file.write("     %8.7g\n" %  parms_dict['conversion_constant'] )
    text_file.write("# Fit offset for maximum stress 1 for yes, 0 for no\n" )
    text_file.write("     %i\n" %  1 )
    text_file.write("# Starting offset value, in degrees\n" )
    text_file.write("     %8.4g\n" %  10.0000 )
    text_file.write("# Fit beam center; 1 for yes, 0 for no\n" )
    text_file.write("     %i\n" %  0 )
    text_file.write("# Material properties set (1/0)\n" )
    text_file.write("     %i\n" %  1 )
    text_file.write("# Peaks properties set (1/0)\n" )
    text_file.write("     %i\n" %  1 )
    
    #write number of peaks and hkls.
    numpeaks = 0
    for x in range(len(FitSettings.fit_orders)):
        numpeaks = numpeaks + len(FitSettings.fit_orders[x]['peak'])
    text_file.write("# Number of peaks\n" )
    text_file.write("     %i\n" %  numpeaks )
    text_file.write("# Peaks info (use, h, k, l)\n" )
    for x in range(len(FitSettings.fit_orders)):
        for y in range(len(FitSettings.fit_orders[x]['peak'])):
            if 'hkl' in FitSettings.fit_orders[x]['peak'][y]:
                hkl = str(FitSettings.fit_orders[x]['peak'][y]['hkl'])
                
                if hkl == '0' or hkl == 0:
                    hkl = '000'
                    
                #check if the d-spacing fits are NaN or not. if NaN switch off (0) otherwise use (1).
                # Got first file name        
                filename = os.path.splitext(os.path.basename(diff_files[0]))[0]
                filename = filename+'.json'
                # Read JSON data from file
                with open(filename) as json_data:
                    fit = json.load(json_data)
                #check if the d-spacing fits are NaN or not. if NaN switch off.
                if np.isnan(fit[x]['peak'][y]['d-space'][0]): 
                    use = 0
                else:
                    use = 1
                    
                #check if the  phase name matches the material (if exists)
                if 'Phase' or 'Output_ElasticProperties' in FitParameters: # FIX ME: should catch differences between output elastic properties and phases. 
                    if FitSettings.phase == FitSettings.fit_orders[x]['peak'][y]['phase']:
                        use = use
                    else:
                        use = 0
            else:
                hkl = '000'
                use = 0
            pos = 0
            if hkl[0] == '-':
                h = hkl[pos:pos+2]
                pos = pos+2
            else:
                h = hkl[pos:pos+1]
                pos = pos+1
            if hkl[pos] == '-':
                k = hkl[pos:pos+2]
                pos = pos+2
            else:
                k = hkl[pos:pos+1]
                pos = pos+1
            if hkl[pos] == '-':
                l = hkl[pos:pos+2]
                pos = pos+2
            else:
                l = hkl[pos:pos+1]
                pos = pos+1
            text_file.write(" %5i    %s    %s    %s\n" % (use, h, k, l))
            
    #material properties
    text_file.write("# Material properties\n" )
    if 'Material' in FitParameters:
        
        raise ValueError("''Material'' is depreciated as an option. Change your input file to have a ''Phase'' instead.")
        
    elif 'Output_ElasticProperties' in FitParameters:
        
        fid = open(FitSettings.Output_ElasticProperties, 'r')
        
        #pipe ealstic properties to the output file.
        text_file.write(fid.read())
    
    elif 'phase' in FitParameters:
        
        try: 
            fid = open((FitSettings.phase + '_elastic_properties.txt'), 'r')
            #pipe ealstic properties to the output file.
            text_file.write(fid.read())
        except:
            raise ValueError("The file " + FitSettings.phase + "_elastic_properties.txt is not found. Edit input file and try again.")
            
    else:
        
        # FIX ME: here I am just writing the elastic properties of BCC iron with no regard for the structure of the file or the avaliable data. 
        #         This should really be fixed. 
        text_file.write("# Name:  (N.B. this is the default phase)\n")
        text_file.write("BCC-Fe\n")
        text_file.write("# Symmetry\n")
        text_file.write("cubic  \n")
        text_file.write("# EOS stuff (v0, k0, k'0)\n")
        text_file.write("      23.5530      159.900      6.52000\n")
        text_file.write("# Elastic model\n")
        text_file.write("       1\n")
        text_file.write("# Parameters for isotropic elastic model (k0 k1 k2, g0, g1, g2)\n")
        text_file.write("      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000\n")
        text_file.write("# Parameters for anisotropic elastic model (Cij)\n")
        text_file.write("      223.000      127.000      127.000      0.00000      0.00000      0.00000\n")
        text_file.write("      127.000      223.000      127.000      0.00000      0.00000      0.00000\n")
        text_file.write("      127.000      127.000      223.000      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      122.000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      122.000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      122.000\n")
        text_file.write("# Parameters for anisotropic elastic model (dCij/dp)\n")
        text_file.write("      6.12245      4.08163      4.08163      0.00000      0.00000      0.00000\n")
        text_file.write("      4.08163      6.12245      4.08163      0.00000      0.00000      0.00000\n")
        text_file.write("      4.08163      4.08163      6.12245      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      2.44898      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      2.44898      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      2.44898\n")
        text_file.write("# Parameters for anisotropic elastic model (d2Cij/dp2)\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      0.00000\n")
        text_file.write("      0.00000      0.00000      0.00000      0.00000      0.00000      0.00000\n")
        text_file.write("     0.00000      0.00000      0.00000      0.00000      0.00000      0.00000\n")

        
    text_file.close()
   
   


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = ['Requirements', 'ImportImage', 'DetectorCheck', 'Conversion', 'GetMask', 'GetCalibration', 'GetTth',
           'GetAzm', 'GetDsp']


import sys
import os
import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt
from cpf import Med


# For MED files from Energy dispersive diffraction patterns at X17B2, 6-BMB and others.

# For ease of use the *.med files are imported using Mark Rivers' Med.py and Mca.py functions.
# These were downloaded from ??? in June 2019.
# and edited to remove unnecessary dependencies e.g. 'Numeric' and 'LinearAlgebra' packages and functions that had
# further dependencies.
# I don't know how much of this is required but for now I am just adding the packages.

# FIX ME: Need a dependency check.




def Requirements():
    
    RequiredParams = [
            'Calib_param'
            ]
    
    return RequiredParams


def ImportImage(ImageName, debug=False):

    filename, file_extension = os.path.splitext(ImageName)
    
    if file_extension == '.med':
        dat = Med.Med()
        dat.read_file(ImageName)
        im_all = []
        for x in range(dat.n_detectors):
            im_all.append(dat.mcas[x].get_data())
        
    else:
        sys.exit('Unknown file type. A *.med file is required')
    
    return im_all


    
def DetectorCheck(ImageName, detector=None):
    
    
    # FIX ME: we should try importing the image and reading the detector type from it.
    if detector == None:
        
        sys.exit('\n\nDioptas requires a detector type.\nTo list the possible detector types in the command line type:\n   import pyFAI\n   pyFAI.detectors.Detector.registry\n\n')
    # FIX ME: check the detector type is valid.

    return detector


def Conversion(E_in, calib, reverse=0, azm=None):
    #convert energy into d-spacing.
    if azm is not None:
        E_in = np.array([[azm, E_in]]) #FIX ME: this is a horrible way to make the conversion work but it is needed for processing the guesses of MED detectors.
    else:
        E_in = np.array([[0, E_in]])
    
    # print('E_in',E_in, 'calib,', calib)
    # print('type E_in', type(E_in))
    # print(E_in[:,1])
    #Convert Energy (keV) to wavelength (Angstroms)
    # E = hc/lamda;  h = 4.135667662(25)×10−15 eV s; c = 299792458 m/s
    wavelength = 4.135667662e-15*(299792458*1E10)/(E_in[:,1]*1000)
    
    # determine which detector calibration goes with each energy.
    sort_idx = np.array(calib['azimuths']).argsort()
    de = sort_idx[np.searchsorted(calib['azimuths'],E_in[:,0],sorter = sort_idx)]

    if not reverse:
        #convert energy to d-spacing
        dspc_out = []
        for i in range(len(de)):
            dspc_out.append(wavelength[i]/2/np.sin(np.radians(calib['calibs'].mcas[i].calibration.two_theta/2)))
        dspc_out = np.squeeze(np.array(dspc_out))
    else:
        #convert dspacing to energy.
        #N.B. this is the reverse finction so that labels tth and dspacing are not correct. 
        dspc_out = []
        for i in range(len(de)):
            
            dspc_out.append(wavelength[i]/2/np.sin(np.radians(calib['calibs'].mcas[i].calibration.two_theta/2)))
            #dspc_out.append(2*np.degrees(np.arcsin(wavelength/2/calib['calibs'].mcas[i].calibration.two_theta)))
            print('dspc_out', dspc_out)
        dspc_out = np.array(dspc_out)
        
    return dspc_out



def GetMask(MSKfile, ImInts, ImTTH, ImAzi, Imy, Imx, debug=False):
    # Masks for 10 element detector are either for full detector or for an energy range. 
    # Currently this just works by removing  single detector.
    
    ImMsk = MSKfile

    ImMsk = ma.zeros(ImInts.shape)
    for x in range(len(MSKfile)):
        ImMsk[MSKfile[x]-1] = 1
    
    if debug:
        #Plot mask.
        #This is left in here for debugging.
        fig = plt.figure()
        ax = fig.add_subplot(1,2,1)
        plt.subplot(121)
        plt.scatter(ImTTH, ImAzi, s=4, c=ImInts, edgecolors='none', cmap=plt.cm.jet)
        ax.set_ylim([0,360])
        ax = fig.add_subplot(1,2,2)
        plt.subplot(122)
        plt.scatter(ma.array(ImTTH,mask=ImMsk), ma.array(ImAzi,mask=ImMsk), s=4, c=ImInts, edgecolors='none', cmap=plt.cm.jet)
        ax.set_ylim([0,360])
        plt.colorbar()
        plt.show()
    
        plt.close()
    
    
    ImInts = ma.array(ImInts, mask=ImMsk)
    
    
    return ImInts




def GetCalibration(filenam):
    'Parse inputs file, create type specific inputs.'

    parms_dict = {}
    dat = Med.Med(file=filenam)
    parms_dict['DispersionType']   = 'EnergyDispersive'
    parms_dict['calibs'] = dat
    if len(dat.mcas) == 10:
        az = np.linspace(0,180,9)
        az = np.append(az, 270)
        az[0] = az[0]+2
        az[8] = az[8]-2
        parms_dict['azimuths'] = az        
    else:
        sys.exit('\n\nDetector is unknown. Edit function to add number of detectors and associated azimuths.\n\n')

    # 2 theta angle in degrees
    #parms_dict['conversion_constant'] = 6.5
    all_angle = []
    for x in range(parms_dict['calibs'].n_detectors):
        all_angle.append(parms_dict['calibs'].mcas[x].calibration.two_theta)
    parms_dict['conversion_constant'] = all_angle
    #print(parms_dict['conversion_constant'])
        
    return parms_dict

    
def bins(azimu, orders):
    # determine bins to use in initial fitting.
    # assign each data to a chunk corresponding to its azimuth value
    # Returns array with indices for each bin and array of bin centroids
    
    bin_vals = np.unique(azimu)
    
    chunks = []
    azichunks = []
    
    tempazi = azimu.flatten()
    for i in range(len(bin_vals)):
        azichunk = np.where((tempazi == bin_vals[i]))        
        chunks.append(azichunk)

    return chunks, bin_vals


def GetTth(cali_file, parms_dict, pix=None, det=None):
    'Called get two theta but it is really getting the energy.'
    
    channels = np.arange(parms_dict['calibs'].mcas[0].data.shape[0])
    en = []
    for x in range(parms_dict['calibs'].n_detectors):
        en.append(parms_dict['calibs'].mcas[x].channel_to_energy(channels))
        
    return en



def GetAzm(cali_file, parms_dict, pix=None, det=None):
    'Give azimuth value for detector x,y position; have to assume geometry of the detectors'
    
    l = parms_dict['calibs'].mcas[0].data.shape
    azi = np.tile(parms_dict['azimuths'], [l[0],1]).T
    
    return azi



def GetDsp(cali_file, parms_dict, pix=None, det=None):
    'Give d-spacing value for detector x,y position; calibration info in data'
        
    channels = np.arange(parms_dict['calibs'].mcas[0].data.shape[0])
    dspc = []
    for x in range(parms_dict['calibs'].n_detectors):
        dspc.append(parms_dict['calibs'].mcas[x].channel_to_d(channels))
        
    return dspc




__all__ = ['Requirements', 'WriteOutput']


import os

import numpy as np
import cpf.PeakFunctions as ff
import json
import re
import datetime

def Requirements():
    #List non-universally required parameters for writing this output type.
    
    RequiredParams = [
            #'apparently none!
            ]
    OptionalParams = [
            'ElasticProperties', #FIX ME: this needs to be included
            'tc', # which thermocoule to include from 6BMB/X17B2 collection system. default to 1. #FIX ME: this needs to be included
            'Output_directory' # if no direcrtory is specified write to current directory.
            ]
    
    return RequiredParams
    
    
def WriteOutput(FitSettings, parms_dict):
# writes *.exp files required by polydefixED.
# N.B. this is a different file than that required by polydefix for monochromatic diffraction.
# This file contains a list of all the diffraction information. Hence it has to be written after the fitting as a single operation.


    FitParameters = dir(FitSettings)
    
    #Parse the required inputs. 
    base_file_name = FitSettings.datafile_Basename
    
    # diffraction patterns 
    # Identical to code in XRD_FitPatterns
    if not 'datafile_Files' in FitParameters:
        n_diff_files = FitSettings.datafile_EndNum - FitSettings.datafile_StartNum + 1
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(j+FitSettings.datafile_StartNum).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    elif 'datafile_Files' in FitParameters:
        n_diff_files = len(FitSettings.datafile_Files)
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(FitSettings.datafile_Files[j]).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    else:
        n_diff_files = len(FitSettings.datafile_Files)

    if 'Output_directory' in FitParameters:
        out_dir = FitSettings.Output_directory
    else:
        out_dir = './'
            

    # create output file name from passed name
    path, filename = os.path.split(base_file_name)
    base, ext = os.path.splitext(filename)
    if base[-1:] == '_': #if the base file name ends in an '_' remove it. 
        base = base[0:-1]
    out_file = out_dir + base + '.exp'

    text_file = open(out_file, "w")
    print('Writing', out_file)

    #headers. set file version to be 1.
    text_file.write("# Experiment analysis file. to be used with PolydefixED\n")
    text_file.write("# For more information: http://merkel.zoneo.net/Polydefix/\n" )
    text_file.write("# File Created by FPF_WritePolydefixED function in FourierPeakFit\n")
    text_file.write("# For more information: http://www.github.com/me/something\n" )
    text_file.write("# File version\n" )
    text_file.write("     1\n" )


    # write two theta angle.
    # Writes two theta from the first detector.
    # FIX ME: should this be the average of all the two theat angles?
    text_file.write("# 2 theta angle (in degrees)\n")
    text_file.write("%12.5f\n" % parms_dict['calibs'].mcas[0].calibration.two_theta)                
       
    # Write detector properties.             
    text_file.write("# Number of detector positions\n")
    text_file.write("%6i\n" % len(parms_dict['azimuths']))
    text_file.write("# Angles for detector positions: Number. Use (1/0). Angle. \n")
    for x in range(len(parms_dict['azimuths'])):
        
        # determine if detector masked
        if FitSettings.Calib_mask:
            if x+1 in FitSettings.Calib_mask:
                use = 0
            else:
                use = 1
                
        # print detector number, if it is being used, then angle, lastly just a number.
        text_file.write("%6i  %i  %7.3f  %i \n" % (x+1, use, parms_dict['azimuths'][x], 1))
        
        # FIX ME: if 10 element detector then turn detector 10 off.
        
        
    #Write number of peaks
    num_subpatterns = len(FitSettings.fit_orders)
    num_peaks = 0
    for x in range(num_subpatterns):
        num_peaks = num_peaks + len(FitSettings.fit_orders[x]['peak'])
    text_file.write("# Number of peaks\n")
    text_file.write("%6i \n" % num_peaks)
    
    #Write peak properties
    text_file.write("# Peak information \n")
    text_file.write("# Number. use (1/0). h. k. l\n")
    peak = 1
    for x in range(num_subpatterns):
        for y in range(len(FitSettings.fit_orders[x]['peak'])):
            if 'hkl' in FitSettings.fit_orders[x]['peak'][y]:
                hkl = str(FitSettings.fit_orders[x]['peak'][y]['hkl'])
                
                #check if the d-spacing fits are NaN or not. if NaN switch off (0) otherwise use (1).
                # Got first file name        
                filename = os.path.splitext(os.path.basename(diff_files[0]))[0]
                filename = filename+'.json'
                # Read JSON data from file
                with open(filename) as json_data:
                    fit = json.load(json_data)
                #check if the d-spacing fits are NaN or not. if NaN switch off.
                if np.isnan(fit[x]['peak'][y]['d-space'][0]): 
                    use = 0
                else:
                    use = 1
            else:
                hkl = '000'
                use = 0
            pos = 0
            if hkl[0] == '-':
                h = hkl[pos:pos+2]
                pos = pos+2
            else:
                h = hkl[pos:pos+1]
                pos = pos+1
            if hkl[pos] == '-':
                k = hkl[pos:pos+2]
                pos = pos+2
            else:
                k = hkl[pos:pos+1]
                pos = pos+1
            if hkl[pos] == '-':
                l = hkl[pos:pos+2]
                pos = pos+2
            else:
                l = hkl[pos:pos+1]
                pos = pos+1
            text_file.write(" %5i %5i    %s    %s    %s\n" % (peak, use, h, k, l))
            peak = peak+1
    
    # Write fetting behaviour
    text_file.write("# Fit offset for maximum stress 1 for yes. 0 for no\n")
    text_file.write("     1\n")
    text_file.write("# Starting offset value. in degrees\n")
    text_file.write("%6i \n" % parms_dict['azimuths'][0])
    
    #Write material properties
    # FIX ME: need to be able to import material properties and write them to the file without messing about.
    text_file.write("# Material properties set (1/0)\n")
    text_file.write("     1\n")
    text_file.write("# Material properties\n")
    text_file.write("# Version\n")
    text_file.write("2       \n")
        
    if 'Material' in FitParameters:
        
        raise ValueError("''Material'' is depreciated as an option. Change your input file to have a ''Phase'' instead.")
        
    elif 'Phase' in FitParameters:
        #left empty on purpose
        text_file.write("")
        
    elif 'Output_ElasticProperties' in FitParameters:
        
        fid = open(FitSettings.Output_ElasticProperties, 'r')
        
        #pipe ealstic properties to the output file.
        text_file.write(fid.read())
        
    else:
        
        # FIX ME: here I am just writing the elastic properties of olivine with no regard for the structure of the file or the avaliable data. 
        #         This should really be fixed. 
        text_file.write("# Name\n")
        text_file.write("Olivine\n")
        text_file.write("# Symmetry\n")
        text_file.write("ortho  \n")
        text_file.write("# EOS stuff (v0. k0. dk0/P. dK0/dT) (Li Li2006)\n")
        text_file.write("# Thermal expansion coefficients (a. b. c)\n")
        text_file.write("# 2.69000e-05 2.12000e-08 0.00000\n")
        text_file.write("# EOS stuff (v0. k0. dk0/P. dK0/dT)       \n")
        text_file.write(" 292.58 128.800 4.2000 0.00000         \n")
        text_file.write("# Thermal expansion coefficients (a. b. c) (ISAAK ET AL.. 1989)   \n")
        text_file.write(" 3.40e-05 0.7100e-08 0.00000          \n")
        text_file.write("# Elastic model           \n")
        text_file.write("1             \n")
        text_file.write("# Parameters for isotropic elastic model (g0. g1. g2. g3. g4)   \n")
        text_file.write("# G = G0 + G1*P + G2*P*P + G3*(T-300) + G4*(T-300)*(T-300)  \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000        \n")
        text_file.write("# Parameters for anisotropic elastic model (Cij) (ISAAK ET AL.. 1989)  \n")
        text_file.write(" 323.700 66.4000 71.6000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 66.4000 197.600 67.2300 0.00000 0.00000 0.00000       \n")
        text_file.write(" 71.6000 67.2300 235.100 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 64.6200 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 78.0500 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 79.0400       \n")
        text_file.write("# Parameters for anisotropic elastic model (dCij/dp) (ZHA. 1996)     \n")
        text_file.write(" 7.98000 4.74000 4.48000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 4.74000 6.37000 3.76000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 4.48000 3.76000 6.38000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 2.17000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 1.64000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 2.31000       \n")
        text_file.write("# Parameters for anisotropic elastic model (d2Cij/dp2)       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write("# Parameters for anisotropic elastic model (dCij/dT) (ISAAK ET AL.. 1989)   \n")
        text_file.write(" -0.0340000 -0.0105000 -0.00940000 0.00000 0.00000 0.00000       \n")
        text_file.write(" -0.0105000 -0.0285000 -0.00510000 0.00000 0.00000 0.00000       \n")
        text_file.write(" -0.00940000 -0.00510000 -0.0286000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 -0.0128000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 -0.0130000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 -0.0157000       \n")
        text_file.write("# Parameters for anisotropic elastic model (d2Cij/dT2)       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000       \n")
        text_file.write(" 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000      \n")
    
    
    # Write details of experiment.
    
    # write number of steps
    text_file.write("# Number of time steps in the experiment\n")
    text_file.write("%i \n" %n_diff_files)
    text_file.write("# Information on time step       \n")
    text_file.write("# Step number.  Step name.         Step time.    Step temperature (K).    taux de def E\n")
    for x in range(n_diff_files):
        temp = np.array(0.)
        with open(diff_files[x], 'r') as myfile:
            for line in myfile:  # For each line, stored as line,
            
                #get date.
                if "DATE:" in line: 
                    ftime = datetime.datetime.strptime(line,"DATE:  %b %d, %Y %H:%M:%S.%f ")
                    if x==0:
                        t_o = ftime
                    t = ftime - t_o #time relative to first diffraction 
                        
                #get temperature
                if not 'Output_tc' in FitParameters:
                    st = "LVP_tc1_calcs.I"
                else:
                    st = "LVP_tc" + FitSettings.Output_tc + "_calcs.I"
                if st in line:
                    temp = re.findall(r'"(.*?)"', line)
                    temp = float(temp[0]) + 273.
            
        #strip directory and ending from filename
        diff_name = os.path.splitext(os.path.basename(diff_files[x]))[0]
        
        text_file.write("%8i        %s  %8.8g            %6.1f\n" %  (x+1, "{:<18}".format(diff_name), t.seconds, temp))
        

    #Write the experimental data fits.
    text_file.write("# Experimental data \n")
    text_file.write("# Peak number.  h.  k.  l.  d-spacing.  intensity.  detector number. step number E t \n")
    for z in range(n_diff_files):
        
        filename = os.path.splitext(os.path.basename(diff_files[z]))[0]
        filename = filename+'.json'
        
        # Read JSON data from file
        with open(filename) as json_data:
            fit = json.load(json_data)
        
        peak = 1
        for x in range(num_subpatterns):
            for y in range(len(FitSettings.fit_orders[x]['peak'])):
                
                # Write the following structure to the data file.    
                #[peak number     H     K     L     d-spacing     Intensity     detector number     step number   ]  repeat for the number of 
                #...                                                                                ]  peaks*number detectors*number patterns
                #[peak number     H     K     L     d-spacing     Intensity     detector number     step number   ]  in data set.

                if 'hkl' in FitSettings.fit_orders[x]['peak'][y]:
                    hkl = str(FitSettings.fit_orders[x]['peak'][y]['hkl'])
                    use = 1
                else:
                    hkl = '000'
                    use = 0
                pos = 0
                if hkl[0] == '-':
                    h = hkl[pos:pos+2]
                    pos = pos+2
                else:
                    h = hkl[pos:pos+1]
                    pos = pos+1
                if hkl[pos] == '-':
                    k = hkl[pos:pos+2]
                    pos = pos+2
                else:
                    k = hkl[pos:pos+1]
                    pos = pos+1
                if hkl[pos] == '-':
                    l = hkl[pos:pos+2]
                    pos = pos+2
                else:
                    l = hkl[pos:pos+1]
                    pos = pos+1
                                
                if 'symmetry' in FitSettings.fit_orders[x]['peak'][y]:
                    sym = FitSettings.fit_orders[x]['peak'][y]['symmetry']
                else:
                    sym = 1
                    
                for w in range(len(parms_dict['calibs'].mcas)):
                    
                    # determine if detector masked
                    if FitSettings.Calib_mask:
                        if w+1 in FitSettings.Calib_mask:
                            use = 0
                        else:
                            use = 1
                    else:
                            use = 1
                            
                    #only write output for unmasked detectors
                    if use == 1:
                        az = parms_dict['azimuths'][w]
                        
                        peak_d = ff.Fourier_expand((az), fit[x]['peak'][y]['d-space'])
                        #peak_tth = 2.*np.degrees(np.arcsin(wavelength/2/peak_d))
                        peak_i = ff.Fourier_expand((az)*sym, fit[x]['peak'][y]['height'])  #FIX ME - Is this the height of the peak or the integral under it?
                        #peak_w = ff.Fourier_expand((az)*sym, fit['peak'][y]['width'])   #FIX ME - is this the correct half width?
                        #peak_p = ff.Fourier_expand((az)*sym, fit['peak'][y]['profile']) #FIX ME - is this 1 or 0 for Gaussian?
                        text_file.write("%8i        %s   %s   %s   %8.4f  %10.3f       %3i           %3i\n" % (peak, h, k, l, peak_d, peak_i, w+1,  z+1))
                
                peak = peak+1
        
    text_file.close()



__all__ = ['Requirements', 'WriteOutput']

import os

import numpy as np
import cpf.PeakFunctions as ff
import json


def Requirements():
    #List non-universally required parameters for writing this output type.
    
    RequiredParams = [
            'Output_NumAziWrite'
            ]
    OptionalParams = [
            'Output_ElasticProperties', #FIX ME: this needs to be included
            'Output_directory' # if no direcrtory is specified write to current directory.
            ]
    
    return RequiredParams


def WriteOutput(FitSettings, parms_dict):
    #writes output from multifit in the form of *.fit files required for polydefix.
    #writes a separate file for each diffraction pattern.
    #uses the parameters in the json files to do so.

    
    FitParameters = dir(FitSettings)
    
    #Parse the required inputs. 
    base_file_name = FitSettings.datafile_Basename
    
    # diffraction patterns 
    # Identical to code in XRD_FitPatterns
    if not 'datafile_Files' in FitParameters:
        n_diff_files = FitSettings.datafile_EndNum - FitSettings.datafile_StartNum + 1
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(j+FitSettings.datafile_StartNum).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    elif 'datafile_Files' in FitParameters:
        n_diff_files = len(FitSettings.datafile_Files)
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names
    
            #make number of pattern
            n = str(FitSettings.datafile_Files[j]).zfill(FitSettings.datafile_NumDigit)
            
            #append diffraction pattern name and directory
            diff_files.append(FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n + FitSettings.datafile_Ending)
    else:
        n_diff_files = len(FitSettings.datafile_Files)

    if 'Output_directory' in FitParameters:
        out_dir = FitSettings.Output_directory
    else:
        out_dir = './'
        
    #def WriteOutput(base_file_name, data_to_write, Num_Azi, wavelength):
    ## writes *.fit files required by polydefix.

    # force Num_Azi to be a float
    #Num_Azi = float(Num_Azi)
    Num_Azi = FitSettings.Output_NumAziWrite
    wavelength = parms_dict['conversion_constant']


    for z in range(n_diff_files):
        
        #read file to write output for
        filename = os.path.splitext(os.path.basename(diff_files[z]))[0]
        filename = filename+'.json'
                
        # Read JSON data from file
        with open(filename) as json_data:
            data_to_write = json.load(json_data)
        
        
        # create output file name from passed name
        path, filename = os.path.split(diff_files[z])
        base, ext = os.path.splitext(filename)
        out_file = out_dir + base + '.fit'
        
        print('Writing', out_dir + base + '.fit')
    
    
        text_file = open(out_file, "w")


        #headers. set file version to be 1.
        text_file.write("# File version\n" )
        text_file.write("1\n" )
    
        #header peak profile
        # I dont know how used this is.
        #Also need to write a single number.
        if 'profile' in data_to_write:
            to_write = data_to_write['profile']
        else:
            to_write = 0
    
        text_file.write("# Peak profile (0: gauss, 1: pseudo-voigt, 2: lorentz)\n" )
        text_file.write("%8i\n" % to_write)
        # FIX ME: this is currently fixed. It should be read from the input constraints.
        
        #number of subpatterns
        # FIX ME: I dont know how the data will be input here.
        # num_subpatterns = 2
    
        num_subpatterns = len(data_to_write)
    
        text_file.write("# Number of sub-patterns\n" )
        text_file.write("%8i\n" % num_subpatterns)
    
        #print type(num_subpatterns)
        #print num_subpatterns
        
        for j in range(num_subpatterns):
    
            text_file.write("# Sub-patterns        %i\n" % j)
    
            #number of peaks
            text_file.write("# Number of peaks\n")
            text_file.write("%8i\n" % len(data_to_write[j]['peak']))
    
            #peak profile
            text_file.write("# Peak profile\n")
            pro = data_to_write[j]['peak'][0]['profile']
            if pro == 0:
                profile_write = 0
            elif pro == 1:
                profile_write = 2
            else:
                profile_write = 1
            text_file.write("%8i\n" % profile_write)
    
            #width factor
            text_file.write("# Width factor\n")
            text_file.write("%13.5f\n" % 15)
            #[0.000000]
    
            #Range in azimuth of each window/subpattern. Preceeded by the number of azimuths
            text_file.write("# Range in azimuth: nvalues, and values\n")
            #[number of azimuth slices]
            text_file.write("%10i\n" % Num_Azi)
            #[azimuth1
            #...
            #azimuth last]
            for k in range(int(Num_Azi)):
                az = k/Num_Azi*360
                text_file.write("%10.1f\n" % az)
    
            #Start and end of the back ground in each subpattern 
            text_file.write("# background positions, in 2 theta (left/right)\n")
            #[bg1left      bg1right
            #... These are the minimum and maximum extent of the background in 2 theta
            #BGlastLeft    BGlastRight]
            for k in range(int(Num_Azi)):
                text_file.write("%13.4f %13.4f\n" % (data_to_write[j]['range'][0][0], data_to_write[j]['range'][0][1]))
    
            # value of the background intensity and slope (assuming a linear background)
            text_file.write("# background coefficients\n")
            for k in range(int(Num_Azi)):
                az = np.array([k/Num_Azi*360])
                inter = ff.Fourier_expand(az, data_to_write[j]['background'][0])
                if len(data_to_write[j]['background']) > 1:
                    slop = ff.Fourier_expand(az, data_to_write[j]['background'][1])
                else:
                    slop = 0
    
                text_file.write("%13.4f %13.4f\n" % (inter, slop))
    
    
            #[bg1intensity      bg1slope
            #... These are the intensity at the min 2 theta and slope the background against 2 theta      ]
            #BGlastIntensity    BGlastSlope]                                                              ]
    
            #header for the peaks in the file
            text_file.write("# Peak positions (2 theta), intensity, half-width, weight Gauss/Lorentz (for pseudo-voigt),\n")
            # write the peak properties for each azimuth slice. and each peak.
            num_peaks = len(data_to_write[j]['peak'])
            for k in range(num_peaks):
                text_file.write("# peak number        %i\n" % k)
                #[position1     Intensity1  HW1            Weight1            ]  repeat for the
                #...                                                          ]  number of peaks in
                #positionLast  IntensityLast  HWLast      WeightLast]         ]  subpattern
                #print data_to_write[j]['peak'][k]['d-space']
                if 'symmetry' in data_to_write[j]['peak'][k]:
                    sym = data_to_write[j]['peak'][k]['symmetry']
                else:
                    sym = 1
                for l in range(int(Num_Azi)):
                    az = np.array([l/Num_Azi*360])
                    peak_d = ff.Fourier_expand((az), data_to_write[j]['peak'][k]['d-space'])
                    peak_tth = 2.*np.degrees(np.arcsin(wavelength/2/peak_d))
                    peak_i = ff.Fourier_expand((az)*sym, data_to_write[j]['peak'][k]['height'])  #FIX ME - Is this the height of the peak or the integral under it?
                    peak_w = ff.Fourier_expand((az)*sym, data_to_write[j]['peak'][k]['width'])   #FIX ME - is this the correct half width?
                    peak_p = ff.Fourier_expand((az)*sym, data_to_write[j]['peak'][k]['profile']) #FIX ME - is this 1 or 0 for Gaussian?
    
                    text_file.write("%13.4f %13.4f %13.4f %13.4f\n" % (peak_tth, peak_i, peak_w, peak_p))
    
        text_file.close()
    



#test cases for wirting files
test = 0

if test==1:
    Fouriers_to_write = [{'peak': [{
    'profile': 0, 
    'width': [ 0.03296637], 
    'd-space': [  1.45521770e+02,   3.15812130e-03,   3.63949255e-03,   -1.43496474e+02,  -1.43500658e+02], 
    'height': [ -1.61854953e+08,  -7.90551608e+01,  -1.86863466e+02,
     1.84245125e+08,  -5.01600833e+05,   8.75674029e+02,
     3.71825217e+03,  -5.17425137e+06,   3.65458422e+08,
    -3.90556882e+03,  -2.25773282e+04,   1.86844720e+08,
    -2.28359051e+08,   8.22152351e+03,   5.67280685e+04,
    -3.07040812e+08,   9.14094789e+06,  -8.05594617e+03,
    -6.26194032e+04,   1.24648536e+08,  -5.65124620e+06,
     2.94882628e+03,   2.51878566e+04,  -2.16682847e+07,
     2.17678521e+07]}], 
    'background': [4.0],
    'range': [[11.700006, 12.079979], [1.9937644, 2.0582786]]}, 
    {'peak': [{
    'profile': 0, 
    'width': [ 0.06473334], 
    'd-space': [ -1.48371656e+02,   1.32237912e-03,   1.75463473e-03,   1.49805560e+02,   1.49801480e+02], 
    'height': [  2.05724056e+07,  -9.25144706e+00,  -5.38376022e+00,
    -2.10599065e+07,  -6.44172806e+06,   1.50563973e+02,
     9.46366683e+01,  -6.81122757e+05,  -3.99125724e+07,
    -7.06638114e+02,  -4.55423181e+02,  -6.71592503e+06,
     3.26600155e+07,   1.31413614e+03,   9.36273739e+02,
    -1.63116330e+06,   3.30713638e+06,  -1.04632954e+03,
    -9.05402148e+02,   1.44412561e+07,  -1.51098537e+07,
     2.97575768e+02,   3.39659190e+02,  -4.92552756e+06,
     4.92459547e+06]}], 
    'background': [4.0],
    'range': [[16.600002, 17.099998], [1.4110823, 1.4532731]]}, 
     {'peak': [{
     'profile': 0, 
     'width': [ 0.05999371], 
     'd-space': [ -1.38852649e+02,   8.75848195e-04,   1.30742495e-03,  1.40024571e+02,   1.40021371e+02], 
     'height': [ -2.94680617e+05,  -7.51087348e+00,  -5.01955745e-01,
     1.71972262e+05,   3.01886458e+06,   1.47013354e+02,
    -3.97875774e+01,   1.00812116e+06,   1.31668009e+06,
    -8.64958125e+02,   4.82478179e+02,  -3.80495454e+07,
     2.73111970e+06,   2.04088806e+03,  -1.76521792e+03,
     9.43916850e+07,  -4.37268233e+07,  -2.05402403e+03,
     2.45598540e+03,  -8.07779850e+07,   6.04951412e+07,
     7.38560951e+02,  -1.13116306e+03,   2.35504567e+07,
    -2.35403011e+07]}], 
     'background': [4.0],
     'range': [[20.400015, 20.89999], [1.1566441, 1.1846806]]}]

    file_name = "Output1.txt"

    Azimuths = 360.

    WriteMultiFit(file_name, Fouriers_to_write, Azimuths)


    #Write the fits to a temporary file
    TempFilename = open("Fit_previous_JSON.dat", "w")

    # Write a JSON string into the file.
    json_string = json.dumps(Fouriers_to_write, TempFilename, sort_keys=True, indent=4, separators=(',', ': '))
    TempFilename.write(json_string)

    





#!/usr/bin/env python
# -*- coding: utf-8 -*-


__all__ = ['Requirements', 'ImportImage', 'DetectorCheck', 'Conversion', 'GetMask', 'GetCalibration', 'GetTth',
           'GetAzm', 'GetDsp']

import sys
import os
import h5py
import numpy as np
import numpy.ma as ma
from PIL import Image
from PIL.ExifTags import TAGS
import matplotlib.pyplot as plt

from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import fabio


# For Dioptas functions to change
#   -   Load/import data
#   -   Load mask
#   -   Load Calibration

def Requirements():
    
    RequiredParams = [
            'Calib_param',
            'Calib_detector',
            #'Calib_pixels'
            'AziBins',
            ]
    
    return RequiredParams


def ImportImage(ImageName, debug=False):
    # im = Image.open(ImageName) ##always tiff?- no

    filename, file_extension = os.path.splitext(ImageName)

    if file_extension == '.nxs':
        im_all = h5py.File(ImageName, 'r')
        # FIX ME: This assumes that there is only one image in the nxs file. If there are more then it will only read
        # 1.
        im = np.array(im_all['/entry1/instrument/detector/data'])
    else:
        im_all = fabio.open(ImageName)
        im = im_all.data

    # Dioptas flips the images to match the orientations in Plot2D. 
    # Therefore implemented here to be consistent with Dioptas.
    im = np.array(im)[::-1]
    
    if debug:
        imgplot = plt.imshow(im, vmin=0, vmax=2000)
        plt.colorbar()
        plt.show()

    #    info = i._getexif()
    #    for tag, value in info.items():
    #        decoded = TAGS.get(tag, tag)
    #        ret[decoded] = value
    #
    #    print ret
    #    stop

    # the bits of Fabio that might get most of the parameters not saved in calibration
    # print(im_all.DESCRIPTION)
    # print(im_all.factory)
    # print(im_all.header)
    # print(im_all.header_keys)

    return im


def DetectorCheck(ImageName, detector=None):
    # FIX ME: we should try importing the image and reading the detector type from it.
    if detector is None:
        sys.exit('\n\nDioptas requires a detector type.\nTo list the possible detector types in the command line '
                 'type:\n   import pyFAI\n   pyFAI.detectors.Detector.registry\n\n')
    # FIX ME: check the detector type is valid.

    return detector


def Conversion(tth_in, conv, reverse=0, azm=None):
    # convert two theta values into d-spacing.

	#azm is needed to enable compatibility with the enerdispersive detectors
	
    # convert wavelength into angstroms.
    # this is not longer required because it is done in the GetCalibration stage. 
    wavelength = conv['conversion_constant']
    # print(wavelength)

    if not reverse:
        # convert tth to d_spacing
        dspc_out = wavelength / 2 / np.sin(np.radians(tth_in / 2))

    else:
        # convert d-spacing to tth.
        # N.B. this is the reverse function so that labels tth and d_spacing are not correct.
        # print(tth_in)
        dspc_out = 2 * np.degrees(np.arcsin(wavelength / 2 / tth_in))
        # dspc_out = 2*np.degrees(np.arcsin(wavelength/2/tth_in[:,1]))

    return dspc_out


def GetMask(MSKfile, ImInts, ImTTH, ImAzi, Imy, Imx, debug=False):
    # Dioptas mask is compressed Tiff image. 
    # Save and load functions within Dioptas are: load_mask and save_mask in dioptas/model/MaskModel.py

    ImMsk = np.array(Image.open(MSKfile))
    ImInts = ma.array(ImInts, mask=ImMsk)

    if debug:
        # Plot mask.
        # This is left in here for debugging.
        fig_1 = plt.figure()
        ax1 = fig_1.add_subplot(1, 3, 1)
        ax1.scatter(ImTTH, ImAzi, s=1, c=(ImInts.data), edgecolors='none', cmap=plt.cm.jet, vmin=0, vmax=np.percentile(ImInts.flatten(), 98))
        ax1.set_title('All data')
        
        ax2 = fig_1.add_subplot(1, 3, 2)
        ax2.scatter(ImTTH, ImAzi, s=1, c=ImMsk, edgecolors='none', cmap='Greys')
        ax2.set_title('Mask')
        ax2.set_xlim(ax1.get_xlim())
        
        ax3 = fig_1.add_subplot(1, 3, 3)
        ax3.scatter(ImTTH, ImAzi, s=1, c=(ImInts), edgecolors='none', cmap=plt.cm.jet, vmin=0, vmax=np.percentile(ImInts.flatten(), 98))
        ax3.set_title('Masked data')
        ax3.set_xlim(ax1.get_xlim())
        #ax2.colorbar()
        plt.show()

        plt.close()

    #FIX ME: need to validate size of images vs. detector name. Otherwise the mask can be the wrong size 
    #det_size = pyFAI.detectors.ALL_DETECTORS['picam_v1'].MAX_SHAPE
    #FIX ME : this could probably all be done by using the detector class in fabio. 
    
    return ImInts


def GetCalibration(filenam):
    """Parse inputs file, create type specific inputs."""

    parms_file = open(filenam, 'r')

    filelines = parms_file.readlines()
    parms_dict = {}

    for item in filelines:
        newparms = item.strip('\n').split(':', 1)
        parm = newparms[1].strip()

        value = None
        # print parm
        try:
            value = int(parm)
            parms_dict[(str(newparms[0]).lower())] = value
        except ValueError:
            try:
                value = float(parm)
                parms_dict[newparms[0]] = value
            except ValueError:
                if parm.startswith('['):
                    listvals = parm.strip('[').strip(']').split(',')
                    newlist = []
                    for val in listvals:
                        newValue = None
                        try:
                            newValue = int(val)
                            newlist.append(newValue)
                        except ValueError:
                            try:
                                newValue = float(val)
                                newlist.append(newValue)
                            except ValueError:
                                newlist.append(val.replace("'", "").replace(" ", "").replace("\"", ""))
                    parms_dict[newparms[0]] = newlist
                elif parm.startswith('{'):
                    # print parm
                    listvals = parm.strip('{').strip('}').split(',')
                    newdict = {}
                    for keyval in listvals:
                        # print keyval
                        newkey = keyval.split(':')[0].replace("'", "").replace(" ", "").replace("\"", "")
                        val = keyval.split(':')[1]
                        newValue = None
                        try:
                            newValue = int(val)
                            newdict[str(newkey)] = newValue
                        except ValueError:
                            try:
                                newValue = float(val)
                                newdict[str(newkey)] = newValue
                            except ValueError:
                                newdict[str(newkey)] = val.replace("'", "").replace(" ", "").replace("\"", "")
                    parms_dict[newparms[0]] = newdict
                elif not parm:
                    parms_dict[newparms[0]] = ''

                else:
                    parms_dict[newparms[0]] = str(parm)

    #process 'detector_config' into pixel sizes if needed.
    if 'Detector_config' in parms_dict:
        #print(parms_dict['Detector_config'])
        parms_dict['pixelsize1'] = parms_dict['Detector_config']['pixel1']
        parms_dict['pixelsize2'] = parms_dict['Detector_config']['pixel2']
                   
    # get wavelengths in Angstrom
    parms_dict['conversion_constant'] = parms_dict['Wavelength'] * 1E10

    # force all dictionary labels to be lower case -- makes s
    parms_dict = {k.lower(): v for k, v in parms_dict.items()}

    # report type
    parms_dict['DispersionType'] = 'AngleDispersive'

    # print(parms_dict)
    return parms_dict


def bins(azimu, orders):
    # determine bins to use in initial fitting.
    # assign each data to a chunk corresponding to its azimuth value
    # Returns array with indices for each bin and array of bin centroids
    bins = orders['AziBins']
    
    azmax = azimu.max()
    azmin = azimu.min()
    binsize = (azmax - azmin) / bins    
    chunks = []
    azichunks = []

    tempazi = azimu.flatten()
    
    for i in range(bins):
        start = azmin + i * binsize
        end = azmin + (i + 1) * binsize
        azichunk = np.where((tempazi > start) & (tempazi <= end))
        azichunks.append(((end - start) / 2) + start)
        chunks.append(azichunk)

    return chunks, azichunks



def GetTth(cali_file, poni, pix=None, det=None):
    'Give 2-theta value for detector x,y position; calibration info in data'
    ai = AzimuthalIntegrator(poni['distance'], poni['poni1'], poni['poni2'], poni['rot1'], poni['rot2'], poni['rot3'],
                             pixel1=poni['pixelsize1'], pixel2=poni['pixelsize2'], detector=det,
                             wavelength=poni['wavelength'])
    # Tth = ai.twoThetaArray()
    # print('Tth', np.min(Tth)/np.pi*180, np.max(Tth)/np.pi*180)
    return np.degrees(ai.twoThetaArray())


def GetAzm(cali_file, poni, pix=None, det=None):
    'Give azimuth value for detector x,y position; calibration info in data'
    ai = AzimuthalIntegrator(poni['distance'], poni['poni1'], poni['poni2'], poni['rot1'], poni['rot2'], poni['rot3'],
                             pixel1=poni['pixelsize1'], pixel2=poni['pixelsize2'], detector=det,
                             wavelength=poni['wavelength'])
    # Chi = ai.chiArray()
    # print('Chi', np.min(Chi)/np.pi*180, np.max(Chi)/np.pi*180)
    return np.degrees(ai.chiArray())


def GetDsp(cali_file, poni, pix=None, det=None):
    'Give d-spacing value for detector x,y position; calibration info in data'
    'Get as thetheta and then convert into d-spacing'
    ai = AzimuthalIntegrator(poni['distance'], poni['poni1'], poni['poni2'], poni['rot1'], poni['rot2'], poni['rot3'],
                             pixel1=poni['pixelsize1'], pixel2=poni['pixelsize2'], detector=det,
                             wavelength=poni['wavelength'])

    # dspc = poni['wavelength']*1E10/2/np.sin(ai.twoThetaArray()/2)
    # print('dspc', np.min(dspc), np.max(dspc))
    return poni['wavelength'] * 1E10 / 2 / np.sin(ai.twoThetaArray() / 2)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="continuous-peak-fit",
    version="0.0.1",
    author="Simon Hunt & Danielle Fenech",
    author_email="simon.hunt@manchester.ac.uk",
    description="Fits azimuth/time dependency of peaks with Fourier Series descriptions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ExperimentalMineralPhysics/Continuous-Peak-Fit",
    packages=setuptools.find_packages(),
    scripts=['bin/CPF_XRD_FitPattern', 'bin/CPF_generate_inputs'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=["matplotlib",
                      "pycairo",
                      "lmfit",
                      "h5py",
                      "pillow",
                      "pyFAI"
                      ]
)


#!/usr/bin/env python

__all__ = ['execute', 'writeoutput']

import json
import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import importlib.util
from cpf.XRD_FitSubpattern import FitSubpattern

np.set_printoptions(threshold=sys.maxsize)

# Needed for JSON to save fitted parameters. 
# copied from https://stackoverflow.com/questions/3488934/simplejson-and-numpy-array#24375113
# on 13th November 2018
def json_numpy_serialzer(o):
    """ Serialize numpy types for json

    Parameters:
        o (object): any python object which fails to be serialized by json

    Example:

        >>> import json
        >>> a = np.array([1, 2, 3])
        >>> json.dumps(a, default=json_numpy_serializer)

    """
    numpy_types = (
        np.bool_,
        # np.bytes_, -- python `bytes` class is not json serializable     
        # np.complex64,  -- python `complex` class is not json serializable  
        # np.complex128,  -- python `complex` class is not json serializable
        # np.complex256,  -- special handling below
        # np.datetime64,  -- python `datetime.datetime` class is not json serializable
        np.float16,
        np.float32,
        np.float64,
        # np.float128,  -- special handling below
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        # np.object_  -- should already be evaluated as python native
        np.str_,
        np.timedelta64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.void,
    )

    if isinstance(o, np.ndarray):
        return o.tolist()
    elif isinstance(o, numpy_types):
        return o.item()
    elif isinstance(o, np.float128):
        return o.astype(np.float64).item()
    # elif isinstance(o, np.complex256): -- no python native for np.complex256
    #     return o.astype(np.complex128).item() -- python `complex` class is not json serializable 
    else:
        raise TypeError("{} of type {} is not JSON serializable".format(repr(o), type(o)))


def initiate(settings_file=None, inputs=None, outtype=None):    
    """
    :param settings_file:
    :param inputs:
    :return FitParameters:
    :return FitSettings:
    """

    if settings_file:
        try:
            print('  ')
            print('The name of the settings file is: ', settings_file)

            if not str.endswith(settings_file, '.py'):
                settings_file = settings_file + '.py'
            spec = importlib.util.spec_from_file_location("settings_module", settings_file)
            FitSettings = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(FitSettings)

            FitParameters = dir(FitSettings)
            
            FitSettings.inputfile = settings_file
        except:
            raise FileNotFoundError

    elif inputs:
        raise NotImplementedError

    else:
        raise ValueError('No inputs provided')

    # List all the required parameters.
    # NOTE: No doubt this will change hence denoted as a list that is worked over.
    # 1. Have option to have list of files rather than sequence. This list needs to be called diff_files.

    # Inputs:
    #   calibration type. (required)
    #   data - in form of:
    #       - file listing data files. (with or without directory)
    #       - file listing data file numbers (needs directory, file name, number digits, ending)
    #       - beginning and and numbers for files (needs directory, file name, number digits, ending)

    RequiredList = ['Calib_type',
                    # 'Calib_data',        # Removed because this is not strictly required.
                    # 'Calib_param',       # Now added to calibration function file.
                    # 'Calib_pixels',      # this is only needed for GSAS-II and should be read from image file. FIX
                    # ME: add to GSAS-II inputfile.
                    # 'Calib_mask',        # a mask file is not strictly required.
                    'datafile_directory',
                    'datafile_Basename',
                    'datafile_Ending',
                    # 'datafile_StartNum',  # now optionally replaced by datafile_Files
                    # 'datafile_EndNum',    # now optionally replaced by datafile_Files
                    'datafile_NumDigit',
                    #'AziBins',            # required based on detector type
                    'fit_orders',
                    # 'Output_type',		   # should be optional
                    # 'Output_NumAziWrite',  # should be optional
                    # 'Output_directory']	   # should be optional
                    ]
    AlternativesList = [
        [['datafile_StartNum', 'datafile_EndNum'], ['datafile_Files']]
    ]

    # load the detector functions here.
    if not 'Calib_type' in FitParameters:
        raise ValueError(
            'There is no ''Calib_type'' in the settings. The fitting cannot proceed until a recognised '
            'Calibration type is present.')
    else:
        calib_mod = 'cpf.' + FitSettings.Calib_type + 'Functions'
    try:
        det = importlib.import_module(calib_mod)
    except ImportError:
        raise ImportError(
            'The ''Calib_type'' ' + FitSettings.Calib_type + ' is not recognised; the file ' + calib_mod +
            ' does not exist. Check if the calibration type exists.')
        # FIX ME: List possible calibration types.

    # Add detector requirements to the Required List of parameters - which has been read from the Detector functions.
    RequiredList = RequiredList + det.Requirements()

    # load output function(s) and check output options are present and valid
    output_mod = []
    if outtype is not None:
        if isinstance(outtype, str):
            output_mod.append('cpf.Write' + outtype)
        else:
            for x in range(len(outtype)):
                output_mod.append('cpf.Write' + outtype[x])
        
    elif 'Output_type' in FitParameters:
        if isinstance(FitSettings.Output_type, str):
            output_mod.append('cpf.Write' + FitSettings.Output_type)
        else:
            for x in range(len(FitSettings.Output_type)):
                output_mod.append('cpf.Write' + FitSettings.Output_type[x])
        
    else:
        raise ValueError('No output type. Add ''Output_type'' to input file or specify ''outtype'' in command.')
        
    try:
        for x in range(len(output_mod)):
        
            wr = importlib.import_module(output_mod[x])
            RequiredList = RequiredList + wr.Requirements()
            
    except ImportError:
        raise ImportError(
            'The ''Output_type'' ' + FitSettings.Output_type + ' is not recognised; the file ' + output_mod +
            ' does not exist. Check if the calibration type exists.')
            # FIX ME: List possible output types.
    
    # check all the required values are present.
    all_present = 1
    # properties of the data files.
    for x in range(len(RequiredList)):
        if RequiredList[x] in FitParameters:
            print('Got: ', RequiredList[x])
        else:
            print('The settings file requires a parameter called  \'', RequiredList[x], '\'')
            all_present = 0

    possible = [[[], []] * len(AlternativesList)]
    for x in range(len(AlternativesList)):
        for y in range(2):
            for z in range(len(AlternativesList[x][y])):
                if AlternativesList[x][y][z] in FitParameters:
                    possible[x][y].append(1)
                else:
                    possible[x][y].append(0)

    # exit if all parameters are not present
    if all_present == 0:
        sys.exit(
            "The highlighted settings are missing from the input file. Fitting cannot proceed until they are all present.")

    # Detector check: Check if the detector is required and if so if it is present and recognised.
    if 'Calib_detector' in FitParameters:
        det_name = FitSettings.Calib_detector
    else:
        det_name = None
    FitSettings.Calib_detector = det.DetectorCheck(os.path.abspath(FitSettings.Calib_data), det_name)
    
    return FitSettings, FitParameters



def writeoutput(settings_file=None, FitSettings=None, parms_dict=None, outtype=None, det=None):
    
    # if no params_dist then initiate. Check all the output functions are present and valid.
    if parms_dict == None:
        FitSettings, FitParameters = initiate(settings_file, outtype=outtype)
        
    if det==None: 
        # get parms_dict from calibration parameter file #
        #load detector here
        det = importlib.import_module('cpf.' + FitSettings.Calib_type + 'Functions')
        parms_dict = det.GetCalibration(os.path.abspath(FitSettings.Calib_param))
       
    
    output_mod = []
    if outtype is not None:
        if isinstance(outtype, str):
            output_mod.append('cpf.Write' + outtype)
        else:
            for x in range(len(outtype)):
                output_mod.append('cpf.Write' + outtype[x])
        
    elif 'Output_type' in FitParameters:
        if isinstance(FitSettings.Output_type, str):
            output_mod.append('cpf.Write' + FitSettings.Output_type)
        else:
            for x in range(len(FitSettings.Output_type)):
                output_mod.append('cpf.Write' + FitSettings.Output_type[x])
               
    
    for x in range(len(output_mod)):
    
        wr = importlib.import_module(output_mod[x])
        print('\nWrite output file(s) using', output_mod[x])
        wr.WriteOutput(FitSettings, parms_dict)


def execute(settings_file=None, inputs=None, debug=False, refine=True, save_all=False, propagate=True, iterations=1, track=False):
    """
    :param settings_file:
    :param inputs:
    :param debug:
    :param refine:
    :param iterations:
    :return:
    """    
    
    FitSettings, FitParameters = initiate(settings_file, inputs)
    
    #load detector
    calib_mod = 'cpf.' + FitSettings.Calib_type + 'Functions'
    det = importlib.import_module('cpf.' + FitSettings.Calib_type + 'Functions')
    
    # define locally required names.
    temporary_data_file = 'PreviousFit_JSON.dat'

    # calibration parameter file #
    parms_dict = det.GetCalibration(os.path.abspath(FitSettings.Calib_param))

    # diffraction patterns #
    if not 'datafile_Files' in FitParameters:
        n_diff_files = FitSettings.datafile_EndNum - FitSettings.datafile_StartNum + 1
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names

            # make number of pattern
            n = str(j + FitSettings.datafile_StartNum).zfill(FitSettings.datafile_NumDigit)

            # append diffraction pattern name and directory
            diff_files.append(os.path.abspath(
                FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n +
                FitSettings.datafile_Ending))
    elif 'datafile_Files' in FitParameters:
        n_diff_files = len(FitSettings.datafile_Files)
        diff_files = []
        for j in range(n_diff_files):
            # make list of diffraction pattern names

            # make number of pattern
            n = str(FitSettings.datafile_Files[j]).zfill(FitSettings.datafile_NumDigit)

            # append diffraction pattern name and directory
            diff_files.append(os.path.abspath(
                FitSettings.datafile_directory + os.sep + FitSettings.datafile_Basename + n +
                FitSettings.datafile_Ending))
    else:
        n_diff_files = len(FitSettings.datafile_Files)

    # get calibration for image set.
    # get array for azimuth, two theta and d-space the same size as the image.
    azimu = det.GetAzm(os.path.abspath(FitSettings.Calib_data), parms_dict, FitSettings.Calib_pixels,
                       FitSettings.Calib_detector)
    twotheta = det.GetTth(os.path.abspath(FitSettings.Calib_data), parms_dict, FitSettings.Calib_pixels,
                          FitSettings.Calib_detector)
    dspace = det.GetDsp(os.path.abspath(FitSettings.Calib_data), parms_dict, FitSettings.Calib_pixels,
                        FitSettings.Calib_detector)
    # FIX ME: Should be able to send diffraction patter, mask and calibration parameters to a function and get back
    # three masked arrays.

    # load calibration data/image -- to get intensities for mask.
    im = det.ImportImage(os.path.abspath(FitSettings.Calib_data), debug=debug)
    intens = ma.array(im)

    # create mask from mask file if present.
    # if not make all values valid
    if 'Calib_mask' in FitParameters:
        intens = det.GetMask(FitSettings.Calib_mask, intens, twotheta, azimu, os.path.abspath(FitSettings.Calib_data),
                             FitSettings.Calib_pixels, debug=debug)
    else:
        ImMask = np.zeros_like(intens)
        intens = ma.array(intens, mask=ImMask)
        intens = ma.masked_outside(intens, 0, np.inf)

    # apply mask to arrays
    azimu = ma.array(azimu, mask=intens.mask)
    twotheta = ma.array(twotheta, mask=intens.mask)
    dspace = ma.array(dspace, mask=intens.mask)

    # plot calibration file
    if debug:
        fig = plt.figure()
        if FitSettings.Calib_type == 'Med':
            # plt.scatter(twotheta, azimu, s=intens/10, c=(intens), edgecolors='none', cmap=plt.cm.jet, vmin = 0,
            # vmax = 1000)  #Better for energy dispersive data.  #FIX ME: need variable for vmax.
            for x in range(9):
                plt.plot(twotheta[x], intens[x] + 100 * x)  # Better for energy dispersive data. #FIX ME: need variable
                # for vmax.
            plt.xlabel('Energy (keV)')
            plt.ylabel('Intensity')

        else:
            # plt.scatter(twotheta, azimu, s=4, c=(intens), edgecolors='none', cmap=plt.cm.jet, vmin = 0, vmax = 1000)
            # #Better for monochromatic data.             #FIX ME: need variable for vmax.
            plt.scatter(twotheta, azimu, s=4, c=(intens), edgecolors='none', cmap=plt.cm.jet, vmin=0)
            # Better for monochromatic data.             #FIX ME: need variable for vmax.
            plt.xlabel(r'2$\theta$')
            plt.ylabel('Azimuth')
            plt.colorbar()
        plt.title('Calibration data')
        plt.show()
        plt.close()

    # Process the diffraction patterns #

    for j in range(n_diff_files):

        print('Process ', diff_files[j])

        # Get diffraction pattern to process.
        im = det.ImportImage(diff_files[j])

        # get intensity array and mask it
        intens = ma.array(im, mask=azimu.mask)

        # plot input file
        if debug:
            fig = plt.figure()
            if FitSettings.Calib_type == 'Med':
                # plt.scatter(twotheta, azimu, s=intens/10, c=(intens), edgecolors='none', cmap=plt.cm.jet, vmin = 0,
                # vmax = 1000)  #Better for energy dispersive data.  #FIX ME: need variable for vmax.
                for x in range(9):
                    plt.plot(twotheta[x], intens[
                        x] + 100 * x)  # Better for energy dispersive data.  #FIX ME: need variable for vmax.
                plt.xlabel('Energy (keV)')
                plt.ylabel('Intensity')

            else:
                # plt.scatter(twotheta, azimu, s=4, c=(intens), edgecolors='none', cmap=plt.cm.jet, vmin = 0,
                # vmax = 1000)  #Better for monochromatic data.             #FIX ME: need variable for vmax.
                plt.scatter(twotheta, azimu, s=4, c=(intens), edgecolors='none', cmap=plt.cm.jet, vmin=0,
                            vmax=4000)  # FIX ME: need variable for vmax.
                plt.xlabel(r'2$\theta$')
                plt.ylabel('Azimuth')
                plt.colorbar()
            plt.title(os.path.basename(diff_files[j]))
            plt.show()
            plt.close()
            
        # get previous fit (if exists and required)
        if os.path.isfile(temporary_data_file) and propagate==True:
            # Read JSON data from file
            print('Loading previous fit results from %s' % temporary_data_file)
            with open(temporary_data_file) as json_data:
                previous_fit = json.load(json_data)

        # switch to save the first fit in each sequence.
        if j == 0 or save_all == True:
            SaveFigs = 1
        else:
            SaveFigs = 0

        # Pass each sub-pattern to FPF_Fit_Subpattern for fitting in turn.
        # get number of sub-patterns to be fitted
        n_subpats = len(FitSettings.fit_orders)
        Fitted_param = []
        lmfit_models = []

        for i in range(n_subpats):

            tthRange = FitSettings.fit_orders[i]['range'][0]
            
            if hasattr(FitSettings, 'fit_orders'):#FitSettings.fit_orders:
                orders = FitSettings.fit_orders[i]
            if hasattr(FitSettings, 'AziBins'):
                orders['AziBins'] = FitSettings.AziBins
                
            if 'previous_fit' in locals():
                params = previous_fit[i]                    
            else:
                params = []
                
            #Track the position of the peak centroid 
            #FIX ME; This is crude - the range doesnt change width. so cant account for massive change in stress. But does it need to?
            if track is True and 'previous_fit' in locals():
                print('tthRange', tthRange)
                mid=[]
                for k in range(len(params['peak'])):
                    mid.append(params['peak'][k]['d-space'][0])
                cent = det.Conversion(np.sum(mid)/(k+1), parms_dict, reverse=1, azm=None)
                tthRange[0] = tthRange[0] - ((tthRange[1]+tthRange[0])/2) + cent
                tthRange[1] = tthRange[1] - ((tthRange[1]+tthRange[0])/2) + cent
                print('move by:' , ((tthRange[1]+tthRange[0])/2) + cent)
                print('tthRange', tthRange)
                
                # The PeakPositionSelections are only used if the fits are not being propagated
                if 'PeakPositionSelection' in orders:
                    for k in range(len(orders['PeakPositionSelection'])):
                        orders['PeakPositionSelection'][k][2] = orders['PeakPositionSelection'][k][2] - ((tthRange[1]+tthRange[0])/2) + cent

            # get subsections of data to pass
            subpat = np.where((twotheta >= tthRange[0]) & (twotheta <= tthRange[1]))
            twotheta_sub = twotheta[subpat]
            dspacing_sub = dspace[subpat]
            azimu_sub = azimu[subpat]
            intens_sub = intens[subpat]


	        # fit the subpattern
            tmp = FitSubpattern([twotheta_sub, dspacing_sub, parms_dict], azimu_sub, intens_sub, orders, params,
                              DetFuncs=calib_mod, SaveFit=SaveFigs, debug=debug, refine=refine, iterations=iterations, fnam=diff_files[j])
            Fitted_param.append(tmp[0])
            lmfit_models.append(tmp[1])
            
        # write output files
        
        # store the fit parameters information as a JSON file.
        filename = os.path.splitext(os.path.basename(diff_files[j]))[0]
        filename = filename + '.json'
        # print(filename)
        with open(filename, 'w') as TempFile:
            # Write a JSON string into the file.
            json_string = json.dump(Fitted_param, TempFile, sort_keys=True, indent=2, default=json_numpy_serialzer)

        # if propagating the fits write them to a temporary file
        if propagate == True:    
            # print json_string
            with open(temporary_data_file, 'w') as TempFile:
    
                # Write a JSON string into the file.
                json_string = json.dump(Fitted_param, TempFile, sort_keys=True, indent=2, default=json_numpy_serialzer)

    # Write the output files.
    writeoutput(settings_file, FitSettings=FitSettings, parms_dict=parms_dict, det=det, outtype=FitSettings.Output_type)
    


if __name__ == '__main__':

    # Load settings fit settings file.
    sys.path.append(os.getcwd())
    settings_file = sys.argv[1]
    print(settings_file)
    execute(inputs=None, settings_file=settings_file, debug=False, refine=True, save_all=False, iterations=1)



"""
Support for Multi-Element Detectors (Med).

Author:        Mark Rivers
Created:       Sept. 18, 2002.  Based on earlier IDL code.
Modifications:
    Edited by SAH (June 2019) to suit analysis code. 
    Stripped out parts that are not required:
        - spline function and option to align the called data. 
"""
from cpf import Mca
import numpy as Numeric
#import spline

#########################################################################
class Med(Mca.Mca):
   """
   The MED class is basically a collection of Mca objects.
   
   This class is device-independent.
   
   Its methods generally simply apply the Mca class methods to each Mca object
   in the collection. The Med class itself is most commonly used for reading
   data from disk files. More importantly, this class is the superclass of the
   epicsMed class.
   """

   def __init__(self, n_detectors=16, file=None):
      """
      Initialization code for creating a new Med object.

      Keywords:
         n_detectors:
            The number of detectors (Mca objects) in the Med.

         file:
            The name of a disk file to read into the Med after it is created.
            The number of detectors in the Med will be changed if the number of
            Mca objects in the disk file is different from the Med.
      """
      Mca.Mca.__init__(self)  # Invoke base class initialization
      self.n_detectors = n_detectors
      self.mcas = []
      for i in range(n_detectors):
         self.mcas.append(Mca.Mca())
      if (file != None): self.read_file(file)


   #########################################################################
   def initial_calibration(self, energy):
      """
      Performs an initial energy calibration for each Mca in the Med.

      Inputs:
         energy:
            The energy of the largest peak in the spectrum.
            
      See the documentation for Mca.initial_calibration() for more information.
      """
      # Read the data first in case this is a hardware Mca
      junk = self.get_data()
      for mca in self.mcas:
         mca.initial_calibration(energy)

   #########################################################################
   def final_calibration(self, peaks):
      """
      Performs a final energy calibration for each Mca in the Med.

      Inputs:
         peaks:
            A list of McaPeak objects. This list is typically read from a
            disk file with function Mca.read_peaks().
            
      See the documentation for Mca.final_calibration() for more information.
      """
      # Read the data first in case this is a hardware Mca
      junk = self.get_data()
      for mca in self.mcas:
         mca.final_calibration(peaks)

   #########################################################################
   def get_energy(self):
      """
      Returns a list of energy arrays, one array for each Mca in the Med.
      See the documentation for Mca.get_energy() for more information.
      """
      energy = []
      for mca in self.mcas:
         energy.append(mca.get_energy())
      return energy

   #########################################################################
   def get_mcas(self):
      """
      Returns a list of Mca objects from the Med.
      """
      return self.mcas

   #########################################################################
   def get_calibration(self):
      """
      Returns a list of McaCalibration objects, one for each Mca in the Med.
      """
      calibration = []
      for mca in self.mcas:
         calibration.append(mca.get_calibration())
      self.calibration = calibration
      return calibration

   #########################################################################
   def set_calibration(self, calibration):
      """
      This procedure sets the calibration parameters for the Med.
      The calibration information is contained in an object or list of 
      objects of type McaCalibration.
      
      Inputs:
         calibration:
            A single object or a list of objects of type McaCalibration
            containing the calibration parameters for each Mca.
            If a single object is passed then this is written to each Mca.
            If a list of objects is passed then calibration[i] is written to
            Mca[i].
      """
      if (isinstance(calibration, Mca.McaCalibration)):
         for mca in self.mcas:
            mca.set_calibration(calibration)
      else:  # Assume it is a list or tuple
         for i in range(self.n_detectors):
            self.mcas[i].set_calibration(calibration[i])


   #########################################################################
   def get_elapsed(self):
      """
      Returns the elapsed parameters for the Med.
      The elapsed information is contained in a list of structures of type
      McaElapsed.
      
      Outputs:
         Returns a list of structures of type McaElapsed.
         
      Procedure:
         This function simply invokes Mca.get_elapsed for each Mca in the Med
         and stores the results in the returned list.
      """
      elapsed = []
      for mca in self.mcas:
         elapsed.append(mca.get_elapsed())
      return elapsed

   #########################################################################
   def set_elapsed(self, elapsed):
      """
      Sets the elapsed parameters for the Med.
      The elapsed information is contained in an object or list of 
      objects of type McaElapsed.

      Inputs:
         elapsed:
            A single structure or a list of structures of type McaElapsed
            containing the elapsed parameters for each Mca.
            If a single object is passed then this is written to each Mca.
            If a list of objects is passed then elapsed[i] is written to Mca[i].
      """
      if (isinstance(elapsed, Mca.McaElapsed)):
         for mca in self.mcas:
            mca.set_elapsed(elapsed)
      else:  # Assume it is a list or tuple
         for i in range(self.n_detectors):
            self.mcas[i].set_elapsed(elapsed[i])


   #########################################################################
   def get_presets(self):
      """
      Returns the preset parameters for the Med.
      The preset information is contained in a list of objects of type
      McaPresets.
      
      Outputs:
         Returns a list of structures of type McaPresets.
         
      Procedure:
         This function simply invokes Mca.get_presets() for each Mca in the Med
         and stores the results in the returned list.
      """
      presets = []
      for mca in self.mcas:
         presets.append(mca.get_presets())
      return presets

   #########################################################################
   def set_presets(self, presets):
      """
      This procedure set the preset parameters for the Med.
      The elapsed information is contained in an object or list of 
      objects of type McaPresets.

      Inputs:
         presets:
            A single object or a list of objects of type McaPresets containing
            the preset parameters for each Mca.
            If a single object is passed then this is written to each Mca.
            If a list of objects is passed then presets[i] is written to Mca[i].
      """
      if (isinstance(presets, Mca.McaPresets)):
         for mca in self.mcas:
            mca.set_presets(presets)
      else:  # Assume it is a list or tuple
         for i in range(self.n_detectors):
            self.mcas[i].set_presets(presets[i])

   #########################################################################
   def get_rois(self):
      """
      Returns the region-of-interest information for each Mca in the Med.

      Outputs:
         Returns a list of list of lists of McaRoi objects.
         The length of the outer list is self.n_detectors, the length of the
         list for each Mca is the number of ROIs defined for that Mca.
      """
      rois = []
      for mca in self.mcas:
         rois.append(mca.get_rois())
      return rois

   #########################################################################
   def get_roi_counts(self, background_width=1):
      """
      Returns the net and total counts for each Roi in each Mca in the Med.

      Outputs:
         Returns a tuple (total, net).  total and net are lists of lists
         containing the total and net counts in each ROI.  The length of the
         outer list is self.n_detectors, the length of the total and net lists
         list for each Mca is the number of ROIs defined for that Mca.
      """
      total = []
      net = []
      for mca in self.mcas:
         t, n = mca.get_roi_counts(background_width)
         total.append(t)
         net.append(n)
      return (total, net)

   #########################################################################
   def set_rois(self, rois, energy=0):
      """
      This procedure sets the ROIs for the Med.
      The elapsed information is contained in a list of McaRoi objects,
      or list of such lists.

      Inputs:
         rois:
            A single list or a nested list of objects McaROI objects.
            If a single list is passed then this is written to each Mca.
            If a list of lists is passed then rois[i][*] is written to Mca[i].
      """
      if (len(rois) <= 1):
         for mca in self.mcas:
            mca.set_rois(rois, energy=energy)
      else:
         for i in range(self.n_detectors):
            self.mcas[i].set_rois(rois[i], energy=energy)

   #########################################################################
   def add_roi(self, roi, energy=0):
      """
      This procedure adds an ROI to each Mca in the Med.

      Inputs:
         roi:
            A single McaROI to be added.
      """
      for mca in self.mcas:
         mca.add_roi(roi, energy=energy)
         
   #########################################################################
   def delete_roi(self, index):
      """
      This procedure deletes the ROI at position "index" from each Mca in the
      Med.

      Inputs:
         index:  The index number of the ROI to be deleted.
      """
      for mca in self.mcas:
         mca.delete_roi(index)

   #########################################################################
   def copy_rois(self, source_mca=0, energy=0):
      """
      This procedure copies the ROIs defined for one Mca in the Med to all of
      the other Mcas.

      Inputs:
         source_mca:
            The index number of the Mca from which the ROIs are to
            be copied.  This number ranges from 0 to self.n_detectors-1.
            The default is the first Mca (index=0).
            
      Keywords:
         energy:
            Set this keyword if the ROIs should be copied by their position
            in energy rather than in channels. This is very useful when 
            copying ROIs when the calibration parameters for each Mca in 
            the Med are not identical.
      """
      rois = self.mcas[source_mca].get_rois(energy=energy)
      self.set_rois(rois, energy=energy)

   #########################################################################
   def get_data(self, total=0, align=0):
      """
      Returns the data from each Mca in the Med as a 2-D Numeric array
      
      Keywords:
         total:
            Set this keyword to return the sum of the spectra from all
            of the Mcas as a 1-D Numeric array.
            
         align:
            Set this keyword to return spectra which have been shifted and
            and stretched to match the energy calibration parameters of the
            first detector.  This permits doing arithmetic on a
            "channel-by-channel" basis. This keyword can be used alone
            or together with the TOTAL keyword, in which case the data
            are aligned before summing.
            
      Outputs:
         By default this function returns a long 2-D array of counts dimensioned
         [nchans, self.n_detectors]
         If the "total" keyword is set then the function returns a long 1-D
         array dimensioned [nchans].
      """
      temp = self.mcas[0].get_data()
      nchans = len(temp)
      data = Numeric.zeros((self.n_detectors, nchans))
      for i in range(self.n_detectors):
         data[i,:] = self.mcas[i].get_data()
      if (align != 0):
          print('This option has been removed')
          stop
#         ref_energy = self.mcas[0].get_energy()
#         for i in range(self.n_detectors):
#            energy = self.mcas[i].get_energy()
#            temp = spline.spline_interpolate(energy, data[i,:], ref_energy)
#            data[i,:] = (temp+.5).astype(Numeric.Int)
      if (total != 0):
         d = Numeric.sum(data)
         self.data = d
         return d
      else:
         self.data = data
         return data

   #########################################################################
   def read_file(self, file, netcdf=0):
      """
      Reads a disk file into an Med object. The file contains the information
      from the Med object which it makes sense to store permanently, but does
      not contain all of the internal state information for the Med.

      Inputs:
         file:
            The name of the disk file to read.
      """
      if (netcdf != 0):
         r = Mca.read_netcdf_file(file)
      else:
         r = Mca.read_ascii_file(file)
      self.name = file
      self.n_detectors = r['n_detectors']
      self.mcas = []
      for i in range(self.n_detectors):
         self.mcas.append(Mca.Mca())
         self.mcas[i].set_rois(r['rois'][i])
         self.mcas[i].set_data(r['data'][i])
         self.mcas[i].set_name(self.name + ':' + str(i+1))
      self.set_elapsed(r['elapsed'])
      self.set_calibration(r['calibration'])
      self.set_environment(r['environment'])



#!/usr/bin/env python

#import sys
#sys.path.append('/Users/simon/g2conda/GSASII/')

__all__ = ['ImportImage', 'CreateGSASIIMask', 'pointInPolygon', 'Fill2ThetaAzimuthMap', 'load_inputs_file', 'LoadMask']

import numpy as np
import numpy.ma as ma
import copy, os, sys
from PIL import Image
import math
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.path as mlp
from scipy.optimize import curve_fit
np.set_printoptions(threshold=sys.maxsize)


#import Plot85 as GsIO

# ------- above here is the intro stuff that I dont know if it is needed. 


def ImportImage(ImageName):


    #Im = GsIO.GetImageData([], ImageName)



    File = open(ImageName,'r')
    save = {}
    #oldThreshold = data['Thresholds'][0]
    S = File.readline()
    while S:
        # if the row is 10 numbers then keep otherise not data.
        if S == 'DATA:':
            S = File.readline()
            continue
        [key,val] = S[:-1].split(':')
        if key in ['Points','Rings','Arcs','Polygons','Frames','Thresholds']:
            #if key == 'Thresholds':
            #    S = File.readline()
            #    continue
            save[key] = eval(val)
            #if key == 'Thresholds':
            #    save[key][0] = oldThreshold
            #    save[key][1][1] = min(oldThreshold[1],save[key][1][1])
        S = File.readline()
    File.close()




    im = Image.open(ImageName) ##always tiff?- no
    # print type(im)
    #im = im.transpose(Image.FLIP_TOP_BOTTOM)
    #im = im.transpose(Image.FLIP_LEFT_RIGHT)
    #print type(im)
    return im

def CreateGSASIIMask(MSKfile, ImInts, ImageSize, ImTTH, ImAzi, Imy,Imx):

    #GSAS-II lisence problems later.
    #As it stands we may have to get people to copy a GSAS-II file from their repository so that it does not
    #have to be distributed by us.
    #GSAS-II lisences give them the right to incorporate wholesale this code into GSAS-II

    msks = LoadMask(MSKfile)

    #make mask array as arrayon ones and zeors for now.
    #Can be transformed into an array-mask later.

    #msk = np.ones((ImageSize[1], ImageSize[2]))
    #ImMsk = np.ones((ImageSize[1], ImageSize[2]))#, dtype=np.int32)
    ImMsk = ImInts

    #Thresholds
    # copied from pyGSAS/GSASIIimage.py Fill2ThetaAzimuthMap
    # units of mask are intensity
    IntLims = msks['Thresholds'][1]
    #print IntLims[0], IntLims[1]
    ImMsk = ma.masked_outside(ImInts,int(IntLims[0]),IntLims[1])
    #ImMsk = ma.masked_outside(ImInts.flatten(),int(IntLims[0]),IntLims[1])


    #Rings
    # copied from pyGSAS/GSASIIimage.py Fill2ThetaAzimuthMap
    # units of mask are two theta and two theta (both in degrees)
    RngLims = (msks['Rings'])
    for twoth,thickness in RngLims:
        #print max(0.01,twoth-thickness/2.), twoth+thickness/2.
        #print type(ma.masked_inside(ImTTH.flatten(),max(0.01,twoth-thickness/2.),twoth+thickness/2.))
        #print type((ImMsk))
        ImMsk.mask = ma.mask_or(ma.getmask(ImMsk),ma.getmask(ma.masked_inside(ImTTH,max(0.01,twoth-thickness/2.),twoth+thickness/2.)))


    #Arcs
    # copied from pyGSAS/GSASIIimage.py Fill2ThetaAzimuthMap
    # units of mask are two theta and azimuth (both in degrees)
    ArcLims = (msks['Arcs'])
    for twoth,azim,thickness in ArcLims:
        tamt = ma.getmask(ma.masked_inside(ImTTH,max(0.01,twoth-thickness/2.),twoth+thickness/2.))
        tama = ma.getmask(ma.masked_inside(ImAzi,azim[0],azim[1]))
        ImMsk.mask = ma.mask_or(ma.getmask(ImMsk),tamt*tama)


    #Points/Spots
    # copied from pyGSAS/GSASIIimage.py Make2ThetaAzimuthMap
    # units of mask are position (x and y) on detector (in mm)
    spots = msks['Points']
    for spX,spY,spdiam in spots:
        tamp = ma.getmask(ma.masked_less((Imx-spX)**2+(Imy-spY)**2,(spdiam/2.)**2))
        ImMsk.mask = ma.mask_or(ma.getmask(ImMsk),tamp)


    #polygon
    # GSAS-II makes a polygon mask in pyGSAS/GSASIIimage.py MakeMaskMap
    # This code though calls a Fortran script.
    # This here is therefore an equivalent code block totally in python.
    # matplotlib.path is needed here.
    # Units of mask are position (x and y) on detector (in mm)
    PolyLms = msks['Polygons']
    points = np.vstack((Imx.flatten(),Imy.flatten())).T
    #FIX ME: The points array is flattened and then in a few lines time grid is reshaped. It is possible to do this without changing the shape of the arrays?
    for polygon in PolyLms:
        path = mlp.Path(polygon)
        grid = path.contains_points(points)
        grid = np.reshape(grid, (Imx.shape[0], Imx.shape[1]))
        ImMsk.mask = ma.mask_or(ma.getmask(ImMsk),grid)


    #frames
    # GSAS-II makes a frame mask in pyGSAS/GSASIIimage.py MakeMaskMap
    # This code though calls a Fortran script.
    # This here is therefore an equivalent code block totally in python.
    # units of mask are position (x and y) on detector (in mm)

    #FIX ME: I think that a frame excludes everything outside the point list, while polgon excludes everything inside the polgon.
    # The difference in the code for the two types is a -TRUE applied to the mask.
    # This should be checked though.

    FrmLms = msks['Frames']
    # print FrmLms
    #for frame in FrmLms:
    #print frame
    if FrmLms:
        path = mlp.Path(FrmLms)
        grid = path.contains_points(points)-True
        print(Imx.shape[0])
        print(Imx.shape[1])
        print(Imx.shape[0]*Imx.shape[1])
        print(grid.shape)
        grid = np.reshape(grid, (Imx.shape[0], Imx.shape[1]))
        ImMsk.mask = ma.mask_or(ma.getmask(ImMsk),grid)


    if 0:
        #This is left in here for debugging.
        fig = plt.figure()
        ax = fig.add_subplot(1,2,1)
        plt.subplot(121)
        plt.scatter(Imx, Imy, s=4, c=intens, edgecolors='none', cmap=plt.cm.jet)
        ax = fig.add_subplot(1,2,2)
        plt.subplot(122)
        plt.scatter(ma.array(Imx,mask=ImMsk.mask), ma.array(Imy,mask=ImMsk.mask), s=4, c=intens, edgecolors='none', cmap=plt.cm.jet)
        plt.colorbar()
        plt.show()

        plt.close()

    return ImMsk


def pointInPolygon(pXY,xy):
    # copied from pyGSAS/GSASIIimage.py
    'Needs a doc string'
    #pXY - assumed closed 1st & last points are duplicates
    Inside = False
    N = len(pXY)
    p1x,p1y = pXY[0]
    for i in range(N+1):
        p2x,p2y = pXY[i%N]
        if (max(p1y,p2y) >= xy[1] > min(p1y,p2y)) and (xy[0] <= max(p1x,p2x)):
            if p1y != p2y:
                xinters = (xy[1]-p1y)*(p2x-p1x)/(p2y-p1y)+p1x
            if p1x == p2x or xy[0] <= xinters:
                Inside = not Inside
        p1x,p1y = p2x,p2y
    return Inside


def Fill2ThetaAzimuthMap(masks, azim, twoth, image):
    'Needs a doc string'
    Zlim = masks['Thresholds'][1]
    rings = masks['Rings']
    arcs = masks['Arcs']
    #TA = np.dstack((ma.getdata(TA[1]),ma.getdata(TA[0]),ma.getdata(TA[2])))    #azimuth, 2-theta, dist
    #tax,tay,tad = np.dsplit(TA,3)    #azimuth, 2-theta, dist**2/d0**2
    for tth,thick in rings:
        tam = ma.mask_or(tam.flatten(),ma.getmask(ma.masked_inside(tay.flatten(),max(0.01,tth-thick/2.),tth+thick/2.)))
    for tth,azm,thick in arcs:
        tamt = ma.getmask(ma.masked_inside(tay.flatten(),max(0.01,tth-thick/2.),tth+thick/2.))
        tama = ma.getmask(ma.masked_inside(tax.flatten(),azm[0],azm[1]))
        tam = ma.mask_or(tam.flatten(),tamt*tama)
    taz = ma.masked_outside(image.flatten(),int(Zlim[0]),Zlim[1])
    tabs = np.ones_like(taz)
    tam = ma.mask_or(tam.flatten(),ma.getmask(taz))
    tax = ma.compressed(ma.array(tax.flatten(),mask=tam))   #azimuth
    tay = ma.compressed(ma.array(tay.flatten(),mask=tam))   #2-theta
    taz = ma.compressed(ma.array(taz.flatten(),mask=tam))   #intensity
    #tad = ma.compressed(ma.array(tad.flatten(),mask=tam))   #dist**2/d0**2
    tabs = ma.compressed(ma.array(tabs.flatten(),mask=tam)) #ones - later used for absorption corr.
    return tax,tay,taz,tabs
    #return tax,tay,taz,tad,tabs

def LoadMask(mskfilename):
    #copied from GSASIIImgGUI.py OnLoadMask

    File = open(mskfilename,'r')
    save = {}
    #oldThreshold = data['Thresholds'][0]
    S = File.readline()
    while S:
        if S[0] == '#':
            S = File.readline()
            continue
        [key,val] = S[:-1].split(':')
        if key in ['Points','Rings','Arcs','Polygons','Frames','Thresholds']:
            #if key == 'Thresholds':
            #    S = File.readline()
            #    continue
            save[key] = eval(val)
            #if key == 'Thresholds':
            #    save[key][0] = oldThreshold
            #    save[key][1][1] = min(oldThreshold[1],save[key][1][1])
        S = File.readline()
    File.close()

    return save


def load_inputs_file(filenam):

    '''
    Parse inputs file, create type specific inputs.
    '''

    filelines = filenam.readlines()
    parms_dict = {}

    for item in filelines:
        newparms = item.strip('\n').split(':',1)
        parm = newparms[1]

        value = None
        # print parm
        try:
            value = int(parm)
            parms_dict[str(newparms[0])] = value
        except ValueError:
            try:
                value = float(parm)
                parms_dict[newparms[0]] = value
            except ValueError:
                if parm.startswith('['):
                    listvals = parm.strip('[').strip(']').split(',')
                    newlist = []
                    for val in listvals:
                        newValue = None
                        try:
                            newValue = int(val)
                            newlist.append(newValue)
                        except ValueError:
                            try:
                                newValue = float(val)
                                newlist.append(newValue)
                            except ValueError:
                                newlist.append(val.replace("'","").replace(" ",""))
                    parms_dict[newparms[0]] = newlist
                elif parm.startswith('{'):
                    # print parm
                    listvals = parm.strip('{').strip('}').split(',')
                    newdict = {}
                    for keyval in listvals:
                        # print keyval
                        newkey = keyval.split(':')[0].replace("'","").replace(" ","")
                        val = keyval.split(':')[1]
                        newValue = None
                        try:
                            newValue = int(val)
                            newdict[str(newkey)] = newValue
                        except ValueError:
                            try:
                                newValue = float(val)
                                newdict[str(newkey)] = newValue
                            except ValueError:
                                newdict[str(newkey)] = val.replace("'","").replace(" ","")
                    parms_dict[newparms[0]] = newdict
                elif not parm:
                    parms_dict[newparms[0]] = ''

                else:
                    parms_dict[newparms[0]] = str(parm)


    return parms_dict




## Functions below taken from GSAS-II code see https://github.com/svaksha/pyGSAS/
## Toby, B. H., & Von Dreele, R. B. (2013). "GSAS-II: the genesis of a modern open-source
## all purpose crystallography software package". Journal of Applied Crystallography,
## 46(2), 544-549. ##

#trig functions
sind = lambda x: math.sin(x*math.pi/180.)
asind = lambda x: 180.*math.asin(x)/math.pi
tand = lambda x: math.tan(x*math.pi/180.)
atand = lambda x: 180.*math.atan(x)/math.pi
atan2d = lambda y,x: 180.*math.atan2(y,x)/math.pi
cosd = lambda x: math.cos(x*math.pi/180.)
acosd = lambda x: 180.*math.acos(x)/math.pi
rdsq2d = lambda x,p: round(1.0/math.sqrt(x),p)
#numpy trig functions
npsind = lambda x: np.sin(x*np.pi/180.)
npasind = lambda x: 180.*np.arcsin(x)/np.pi
npcosd = lambda x: np.cos(x*np.pi/180.)
npacosd = lambda x: 180.*np.arccos(x)/np.pi
nptand = lambda x: np.tan(x*np.pi/180.)
npatand = lambda x: 180.*np.arctan(x)/np.pi
npatan2d = lambda y,x: 180.*np.arctan2(y,x)/np.pi

def makeMat(Angle,Axis):
    '''Make rotation matrix from Angle and Axis
    :param float Angle: in degrees
    :param int Axis: 0 for rotation about x, 1 for about y, etc.
    '''
    cs = npcosd(Angle)
    ss = npsind(Angle)
    M = np.array(([1.,0.,0.],[0.,cs,-ss],[0.,ss,cs]),dtype=np.float32)
    return np.roll(np.roll(M,Axis,axis=0),Axis,axis=1)




def GetTthAzmDsp(x,y,data): #expensive
	## SAH: lifted from: https://subversion.xray.aps.anl.gov/pyGSAS/trunk/GSASIIimage.py
    '''
    '''
    wave = data['wavelength']
    cent = data['center']
    tilt = data['tilt']
    # print tilt, cosd(tilt)
    dist = data['distance']/cosd(tilt)
    x0 = data['distance']*tand(tilt)
    phi = data['rotation']
    dep = data['DetDepth']
    LRazim = data['LRazimuth']
    azmthoff = data['azmthOff']
    dx = np.array(x-cent[0],dtype=np.float32)
    dy = np.array(y-cent[1],dtype=np.float32)
    D = ((dx-x0)**2+dy**2+data['distance']**2)      #sample to pixel distance
    X = np.array(([dx,dy,np.zeros_like(dx)]),dtype=np.float32).T
    # print np.array(([dx,dy,np.zeros_like(dx)]),dtype=np.float32).shape
    X = np.dot(X,makeMat(phi,2))
    Z = np.dot(X,makeMat(tilt,0)).T[2]
    tth = npatand(np.sqrt(dx**2+dy**2-Z**2)/(dist-Z))
    dxy = peneCorr(tth,dep,tilt,npatan2d(dy,dx))
    DX = dist-Z+dxy
    DY = np.sqrt(dx**2+dy**2-Z**2)
    tth = npatan2d(DY,DX)
    dsp = wave/(2.*npsind(tth/2.))
    azm = (npatan2d(dy,dx)+azmthoff+720.)%360.
    G = D/data['distance']**2       #for geometric correction = 1/cos(2theta)^2 if tilt=0.
    return np.array([tth,azm,G,dsp])


def peneCorr(tth,dep,tilt=0.,azm=0.):
    'Needs a doc string'
    return dep*(1.-npcosd(tth))         #best one


#related calls#
def GetTth(x,y,data):
    'Give 2-theta value for detector x,y position; calibration info in data'
    return GetTthAzmDsp(x,y,data)[0]

def GetTthAzm(x,y,data):
    'Give 2-theta, azimuth values for detector x,y position; calibration info in data'
    return GetTthAzmDsp(x,y,data)[0:2]

def GetDsp(x,y,data):
    'Give d-spacing value for detector x,y position; calibration info in data'
    return GetTthAzmDsp(x,y,data)[3]

def GetAzm(x,y,data):
    'Give azimuth value for detector x,y position; calibration info in data'
    return GetTthAzmDsp(x,y,data)[1]



